import functools
from taf.exceptions import RepositoryNotCleanError, RepositoryNotSynced
from taf.git import GitRepository


def check_if_clean_and_synced(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        skip_clean_check = kwargs.pop("skip_clean_check", False)
        skip_remote_check = kwargs.pop("skip_remote_check", False)

        if not skip_clean_check or not skip_remote_check:
            path = kwargs.get("path", None)
            if path is None:
                path = args[0]
            repo = GitRepository(path=path)

        if not skip_clean_check:
            if repo.something_to_commit():
                raise RepositoryNotCleanError(repo.name)

        if not skip_remote_check:
            if repo.has_remote():
                if not repo.synced_with_remote(repo.default_branch):
                    raise RepositoryNotSynced(repo.name)

        return func(*args, **kwargs)

    return wrapper



import json

from taf.auth_repo import AuthenticationRepository
from taf.constants import DEFAULT_RSA_SIGNATURE_SCHEME
from taf.config import load_config
from taf.tuf.keys import get_sslib_key_from_value, _get_legacy_keyid
from taf.log import taf_logger
from pathlib import Path
from typing import Dict, Optional, Union


def find_taf_directory(path: Union[Path, str]) -> Optional[Path]:
    """Look for the .taf directory within the archive root.

    Args:
        path (Path): The current path to do the lookup.

    Returns:
        Optional[Path]: The path to the .taf directory if found, otherwise None.
    """
    # Check the parent directory of the authentication repository
    current_dir = Path(path).absolute()
    while current_dir != current_dir.root:
        taf_directory = current_dir / ".taf"
        if taf_directory.exists() and taf_directory.is_dir():
            return taf_directory
        current_dir = current_dir.parent

    # If not found, check the archive root
    archive_root = Path(path).parent.parent
    current_dir = archive_root
    while current_dir != current_dir.root:
        taf_directory = current_dir / ".taf"
        if taf_directory.exists() and taf_directory.is_dir():
            return taf_directory
        current_dir = current_dir.parent

    taf_logger.debug(f"No .taf directory found starting from {Path(path)}")
    return None


def find_keystore(path: Union[str, Path]) -> Optional[Path]:
    """Find keystore starting from the given path and traversing parent directories if needed."""
    taf_directory = find_taf_directory(Path(path))
    if taf_directory:
        keystore_path = taf_directory / "keystore"
        if keystore_path.exists() and keystore_path.is_dir():
            taf_logger.debug(f"Found keystore at {keystore_path}")
            return keystore_path
    taf_logger.debug(f"No keystore found starting from {path}")
    return None


def _build_keys_name_mapping(raw_mapping: Dict) -> Dict[str, str]:
    """
    Internal helper to turn a raw key_name -> key_data mapping into the
    legacy key_id -> key_name mapping used elsewhere.
    """
    keys_name_mappings: Dict[str, str] = {}

    for key_name, key_data in raw_mapping.items():
        scheme = key_data.get("scheme", DEFAULT_RSA_SIGNATURE_SCHEME)

        if "public" in key_data:
            pub_key = key_data["public"]
            try:
                ssl_pub_key = get_sslib_key_from_value(pub_key, scheme=scheme)
                key_id = _get_legacy_keyid(ssl_pub_key)
                keys_name_mappings[key_id] = key_name
            except ValueError:
                taf_logger.log("NOTICE", f"Invalid public key {key_name}")

        elif "keyid" in key_data:
            key_id = key_data["keyid"]
            keys_name_mappings[key_id] = key_name

    return keys_name_mappings


def read_keys_name_mapping(keys_description: Optional[Union[str, Path]]) -> Dict:
    keys_name_mappings: Dict = {}
    if keys_description is None:
        return keys_name_mappings

    keys_description_path = Path(keys_description)
    if not keys_description_path.is_file():
        return keys_name_mappings

    try:
        keys_description_data = json.loads(keys_description_path.read_text())
    except json.decoder.JSONDecodeError:
        return keys_name_mappings

    key_names_mapping = keys_description_data.get("yubikeys")

    return _build_keys_name_mapping(key_names_mapping) if key_names_mapping else {}


def read_keys_name_mapping_from_auth(auth_repo: AuthenticationRepository) -> Dict:
    """
    Read the keys name mapping from the authentication repository.
    """
    taf_dir = find_taf_directory(auth_repo.path)
    if taf_dir is None:
        return {}

    try:
        cfg = load_config(taf_dir / "config.toml")
        if cfg.root is None:
            raise FileNotFoundError("No root authentication repository found.")
    except FileNotFoundError:
        return {}

    root_auth_repo_name = cfg.root.name
    archive_dir = taf_dir.parent
    root_auth_repo = AuthenticationRepository(path=(archive_dir / root_auth_repo_name))
    if not root_auth_repo.is_git_repository:
        taf_logger.debug(
            f"{root_auth_repo_name} is not a valid authentication repository."
        )
        return {}

    raw_mapping = root_auth_repo.get_keys_mapping() or {}
    return _build_keys_name_mapping(raw_mapping)



"""
config.py - Utility helpers for reading `config.toml` into an *immutable*
`TafConfig` instance

Quick start
-----------
from taf.config import load_config

cfg = load_config()
cfg = load_config("path/to/config.toml")

print(cfg.root.org)                     # -> "some-repo-law-org"
"""

from __future__ import annotations

import importlib
import sys
from pathlib import Path
from typing import Any, Mapping, Optional

import attrs

_BACKEND_NAME = "tomllib" if sys.version_info >= (3, 11) else "tomli"
try:
    toml: Any = importlib.import_module(_BACKEND_NAME)  # type: ignore
except ModuleNotFoundError as exc:
    raise ImportError(
        "Neither the standard-library `tomllib` (Python >= 3.11) or the "
        "third - party `tomli` package is installed - cannot parse TOML."
    ) from exc


@attrs.define(slots=True, frozen=True, kw_only=True)
class RootConfig:
    """Representation of the TOML `[root]` table."""

    name: str
    org: str
    hash: Optional[str]


@attrs.define(slots=True, frozen=True, kw_only=True)
class TafConfig:
    """
    Immutable in-memory representation of `config.toml`.

    Attributes
    ----------
    shallow
        Whether repository is shallow (top-level `shallow = true/false`).
    root
        Nested `[root]` table or *None* if omitted.
    """

    shallow: bool = False
    root: Optional[RootConfig] = None

    @property
    def is_shallow(self) -> bool:
        """Alias for the top-level *shallow* flag (reads nicer as a predicate)."""
        return self.shallow

    @classmethod
    def from_mapping(cls, mapping: Mapping[str, Any]) -> "TafConfig":
        shallow = bool(mapping.get("shallow", False))

        root_tbl = mapping.get("root", {})
        root: Optional[RootConfig] = (
            RootConfig(  # type: ignore[call-arg]
                name=root_tbl["name"],
                org=root_tbl["org"],
                hash=root_tbl.get("hash"),
            )
            if root_tbl
            else None
        )

        return cls(shallow=shallow, root=root)  # type: ignore[call-arg]

    def to_mapping(self) -> dict[str, Any]:
        """Round-trip back to a JSON-serialisable mapping."""
        data: dict[str, Any] = {"shallow": self.shallow}
        if self.root:
            data["root"] = attrs.asdict(self.root)
        return data


def load_config(config_path: str | Path | None = None) -> TafConfig:
    """
    Load *config.toml* and return a `TafConfig` instance.

    Parameters
    ----------
    config_path
        Explicit path to a TOML file.  If *None*, falls back to
        `find_taf_directory() ` / `config.toml`.

    Raises
    ------
    FileNotFoundError
        When the file cannot be found.
    KeyError
        When required TOML keys are missing.
    """
    from taf.api.utils._conf import find_taf_directory  # lazy import to avoid cycles

    base = find_taf_directory(Path.cwd())
    if base is None:
        raise FileNotFoundError("TAF directory not found")

    if config_path is None:
        candidate = base / "config.toml"
    else:
        candidate = Path(config_path).expanduser().resolve()

    if not candidate.is_file():
        raise FileNotFoundError(f"config.toml not found at: {candidate}")

    with candidate.open("rb") as fp:
        data = toml.load(fp)

    return TafConfig.from_mapping(data)



import importlib
import click


class LazyGroup(click.Group):
    def __init__(self, *args, lazy_subcommands=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.lazy_subcommands = lazy_subcommands or {}

    def list_commands(self, ctx):
        return super().list_commands(ctx) + sorted(self.lazy_subcommands.keys())

    def get_command(self, ctx, cmd_name):
        if cmd_name in self.lazy_subcommands:
            return self._lazy_load(cmd_name)
        return super().get_command(ctx, cmd_name)

    def _lazy_load(self, cmd_name):
        function = self.lazy_subcommands[cmd_name]()
        if not function:
            return
        module_name, function_name = function.rsplit(".", 1)
        module = importlib.import_module(module_name)
        function = getattr(module, function_name)
        group = click.Group(name=cmd_name)
        function(group)
        return group

    def format_commands(self, ctx, formatter):
        rows = []
        sub_commands = set(self.lazy_subcommands.keys())
        loaded_commands = set(super().list_commands(ctx))

        for subcommand in sorted(loaded_commands | sub_commands):
            if subcommand in self.lazy_subcommands:
                # Load the command to get its help text
                cmd = self._lazy_load(subcommand)
            else:
                cmd = self.get_command(ctx, subcommand)
            if cmd is None:
                continue

            help = cmd.get_short_help_str()
            rows.append((subcommand, help))

        if rows:
            with formatter.section("Commands"):
                formatter.write_dl(rows)



from datetime import datetime, timezone
from logging import ERROR
from typing import Dict, List, Optional, Tuple
from logdecorator import log_on_error
from taf.api.utils._conf import read_keys_name_mapping
from taf.yubikey.yubikey_manager import PinManager
from tuf.api.metadata import Snapshot, Timestamp

from taf.api.utils._git import check_if_clean_and_synced
from taf.api.api_workflow import manage_repo_and_signers
from taf.exceptions import TAFError
from taf.constants import DEFAULT_RSA_SIGNATURE_SCHEME
from taf.messages import git_commit_message
from taf.tuf.repository import MetadataRepository as TUFRepository
from taf.log import Path, taf_logger
from taf.auth_repo import AuthenticationRepository


@log_on_error(
    ERROR,
    "An error occurred while adding key names: {e}",
    logger=taf_logger,
    on_exceptions=TAFError,
    reraise=False,
)
@check_if_clean_and_synced
def add_key_names(
    path: str,
    keys_description: Path,
    pin_manager: PinManager,
    scheme: Optional[str] = DEFAULT_RSA_SIGNATURE_SCHEME,
    keystore: Optional[str] = None,
    commit: Optional[bool] = True,
    commit_msg: Optional[str] = None,
    prompt_for_keys: Optional[bool] = False,
    push: Optional[bool] = True,
    update_snapshot_and_timestamp: Optional[bool] = True,
):
    if not keys_description.is_file():
        raise TAFError(f"{keys_description} does not exist")

    commit_msg = commit_msg or git_commit_message("add-key-names")

    keys_name_mappings = read_keys_name_mapping(keys_description)
    if not keys_name_mappings:
        raise TAFError(f"{keys_description} is not a valid keys description json file")

    auth_repo = AuthenticationRepository(path=path, pin_manager=pin_manager)
    existing_mapping = auth_repo.keys_name_mappings
    for key_id, key_name in dict(keys_name_mappings).items():
        if key_id in existing_mapping and key_name == keys_name_mappings[key_id]:
            keys_name_mappings.pop(key_id)

    if not len(keys_name_mappings):
        taf_logger.log("NOTICE", "All names already added to the repository")
        return

    auth_repo.add_key_names(keys_name_mappings)
    parent_roles = set()
    for key_id in keys_name_mappings:
        roles_of_key = auth_repo.find_keysid_roles([key_id], check_threshold=False)
        # map the roles to the roles that contain their keys
        for role_of_key in roles_of_key:
            role_containing_key = auth_repo.find_role_containing_key_of_role(
                role_of_key
            )
            if role_containing_key is not None:
                parent_roles.add(role_containing_key)

    with manage_repo_and_signers(
        auth_repo,
        list(parent_roles),
        keystore,
        scheme,
        prompt_for_keys,
        load_snapshot_and_timestamp=update_snapshot_and_timestamp,
        commit=commit,
        commit_msg=commit_msg,
        push=push,
    ):
        auth_repo.set_key_names(keys_name_mappings)
        if update_snapshot_and_timestamp:
            auth_repo.update_snapshot_and_timestamp()


@log_on_error(
    ERROR,
    "An error occurred while checking expiration dates: {e!r}",
    logger=taf_logger,
    on_exceptions=TAFError,
    reraise=False,
)
def check_expiration_dates(
    path: str,
    interval: Optional[int] = 30,
    start_date: Optional[datetime] = None,
    excluded_roles: Optional[List[str]] = None,
    print_output: bool = True,
) -> Tuple[Dict, Dict]:
    """
    Check if any metadata files (roles) are expired or will expire in the next <interval> days.
    Prints a list of expired roles.

    Arguments:
        path: Authentication repository's location.
        interval: Number of days ahead to check for expiration.
        start_date: Date from which to start checking for expiration.
        excluded_roles: List of roles to exclude from the check.

    Side Effects:
        Prints lists of roles that expired or are about to expire.

    Returns:
        None
    """
    taf_repo = TUFRepository(path)

    if start_date is None:
        start_date = datetime.now(timezone.utc)
    elif start_date.tzinfo is None:
        start_date = start_date.replace(tzinfo=timezone.utc)

    expired_dict, will_expire_dict = taf_repo.check_roles_expiration_dates(
        interval, start_date, excluded_roles
    )

    if print_output:
        print_expiration_dates(
            expired_dict, will_expire_dict, start_date=start_date, interval=interval
        )

    return expired_dict, will_expire_dict


def print_expiration_dates(
    expired: Dict, will_expire: Dict, start_date: datetime, interval: Optional[int] = 30
) -> None:
    if expired or will_expire:
        now = datetime.now(timezone.utc)
        print(
            f"Given a {interval} day interval from ({start_date.strftime('%Y-%m-%d')}):"
        )
        for role, expiry_date in expired.items():
            delta = now - expiry_date
            print(
                f"{role} expired {delta.days} days ago (on {expiry_date.strftime('%Y-%m-%d')})"
            )
        for role, expiry_date in will_expire.items():
            delta = expiry_date - now
            print(
                f"{role} will expire in {delta.days} days (on {expiry_date.strftime('%Y-%m-%d')})"
            )
    else:
        print(f"No roles will expire within the given {interval} day interval")


@check_if_clean_and_synced
def update_metadata_expiration_date(
    path: str,
    pin_manager: PinManager,
    roles: List[str],
    interval: Optional[int] = None,
    keystore: Optional[str] = None,
    scheme: Optional[str] = DEFAULT_RSA_SIGNATURE_SCHEME,
    start_date: Optional[datetime] = None,
    commit: Optional[bool] = True,
    commit_msg: Optional[str] = None,
    prompt_for_keys: Optional[bool] = False,
    push: Optional[bool] = True,
    update_snapshot_and_timestamp: Optional[bool] = True,
    keys_description: Optional[str] = None,
) -> None:
    """
    Update expiration dates of the specified roles and all other roles that need
    to be signed in order to guarantee validity of the repository e.g. snapshot
    and timestamp need to be signed after a targets role is updated.

    Arguments:
        path: Authentication repository's location.
        roles: A list of roles whose expiration dates should be updated.
        interval: Number of days added to the start date in order to calculate the
            expiration date.
        keystore (optional): Keystore directory's path
        scheme (optional): Signature scheme.
        start_date (optional): Date to which expiration interval is added.
            Set to today if not specified.
        commit (optional): Indicates if the changes should be committed and pushed automatically.
        commit_msg (optional): Custom commit messages.
        prompt_for_keys (optional): Whether to ask the user to enter their key if it is not located inside the keystore directory.
        push (optional): Flag specifying whether to push to remote

    Side Effects:
        Updates metadata files, saves changes to disk and commits changes
        unless no_commit is set to True.

    Returns:
        None
    """

    auth_repo = AuthenticationRepository(path=path, pin_manager=pin_manager)
    keys_name_mappings = read_keys_name_mapping(keys_description)
    auth_repo.add_key_names(keys_name_mappings)
    if start_date is None:
        start_date = datetime.now()

    commit_msg = commit_msg or git_commit_message(
        "update-expiration-dates", roles=",".join(roles)
    )

    # update the order, snapshot has to be updated before timestamp
    # and all other roles have to be updated before snapshot
    # all other roles can be updated in any order

    update_snapshot_expiration_date = Snapshot.type in roles
    update_timestamp_expiration_date = Timestamp.type in roles

    with manage_repo_and_signers(
        auth_repo,
        roles,
        keystore,
        scheme,
        prompt_for_keys,
        load_snapshot_and_timestamp=update_snapshot_and_timestamp,
        commit=commit,
        commit_msg=commit_msg,
        push=push,
    ):
        if update_snapshot_expiration_date:
            auth_repo.add_to_open_metadata([Snapshot.type])
        if update_timestamp_expiration_date:
            auth_repo.add_to_open_metadata([Timestamp.type])

        for role in roles:
            auth_repo.set_metadata_expiration_date(
                role, start_date=start_date, interval=interval
            )

        auth_repo.remove_from_open_metadata([Snapshot.type])
        # it is important to update snapshot first

        if (update_snapshot_expiration_date or update_snapshot_and_timestamp) and not (
            len(roles) == 1 and update_timestamp_expiration_date
        ):
            auth_repo.do_snapshot(force=True)

        auth_repo.remove_from_open_metadata([Timestamp.type])
        if update_timestamp_expiration_date or update_snapshot_and_timestamp:
            auth_repo.do_timestamp(force=True)


@check_if_clean_and_synced
def update_snapshot_and_timestamp(
    path: str,
    pin_manager: PinManager,
    keystore: Optional[str] = None,
    roles_to_sync: Optional[List[str]] = None,
    scheme: Optional[str] = DEFAULT_RSA_SIGNATURE_SCHEME,
    commit: Optional[bool] = True,
    commit_msg: Optional[str] = None,
    prompt_for_keys: Optional[bool] = False,
    push: Optional[bool] = True,
    update_expiration_dates: Optional[bool] = True,
    keys_description: Optional[str] = None,
) -> None:
    """
    Update expiration snapshot and timestamp

    Arguments:
        path: Authentication repository's location.
        keystore (optional): Keystore directory's path
        scheme (optional): Signature scheme.
        commit (optional): Indicates if the changes should be committed and pushed automatically.
        commit_msg (optional): Custom commit messages.
        prompt_for_keys (optional): Whether to ask the user to enter their key if it is not located inside the keystore directory.
        push (optional): Flag specifying whether to push to remote
        update_expiration_dates (optional): Flag specifying whether to update expiration dates

    Side Effects:
        Updates metadata files, saves changes to disk and commits changes
        unless no_commit is set to True.

    Returns:
        None
    """

    auth_repo = AuthenticationRepository(path=path, pin_manager=pin_manager)
    keys_name_mappings = read_keys_name_mapping(keys_description)
    auth_repo.add_key_names(keys_name_mappings)

    with manage_repo_and_signers(
        auth_repo,
        [],
        keystore,
        scheme,
        prompt_for_keys,
        load_snapshot_and_timestamp=True,
        commit=commit,
        commit_msg=commit_msg,
        push=push,
    ):
        if update_expiration_dates:
            auth_repo.add_to_open_metadata([Snapshot.type, Timestamp.type])
            for role in [Snapshot.type, Timestamp.type]:
                auth_repo.set_metadata_expiration_date(role)
            auth_repo.clear_open_metadata()
        if roles_to_sync:
            auth_repo.sync_snapshot_with_roles(roles_to_sync)
            auth_repo.do_timestamp(force=True)
        else:
            auth_repo.update_snapshot_and_timestamp()



import click
from taf.api.conf import init


def init_command():
    @click.command(help="Create a .taf directory")
    @click.option("--path", default=".", help=".taf directory's location. If not specified, set to the current directory")
    @click.option("--keystore", default=None, help="Location of the keystore directory to copy from. Can be specified in keys-description dictionary")
    @click.option("--keys-description", help="A dictionary containing information about the keys or a path to a json file which stores the needed information")
    def _init_command(path, keystore, keys_description):
        init(path, keystore, keys_description)
    return _init_command


def attach_to_group(group):
    group.add_command(init_command(), name='init')



import platform
import click
import errno
import datetime
import time
import json
import os
import stat
import subprocess
import tempfile
import shutil
import uuid
import sys
from getpass import getpass
from functools import wraps
from pathlib import Path
from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.serialization import (
    load_pem_public_key,
    load_pem_private_key,
)
from json import JSONDecoder
from taf.log import taf_logger
import taf.settings
from taf.exceptions import PINMissmatchError
from typing import List, Optional, Tuple, Dict
from securesystemslib.hash import digest_fileobject
from securesystemslib.storage import FilesystemBackend, StorageBackendInterface


def _iso_parse(date):
    return datetime.datetime.strptime(date, "%Y-%m-%d %H:%M:%S.%f")


class IsoDateParamType(click.ParamType):
    name = "iso_date"

    def convert(self, value, param, ctx):
        if value is None:
            return datetime.datetime.now()

        if isinstance(value, datetime.datetime):
            return value
        try:
            return _iso_parse(value)
        except ValueError as ex:
            self.fail(str(ex), param, ctx)


ISO_DATE_PARAM_TYPE = IsoDateParamType()


def extract_x509(cert_pem):
    cert = x509.load_pem_x509_certificate(cert_pem, default_backend())

    def _get_attr(oid):
        attrs = cert.subject.get_attributes_for_oid(oid)
        return attrs[0].value if len(attrs) > 0 else ""

    return {
        "name": _get_attr(x509.OID_COMMON_NAME),
        "organization": _get_attr(x509.OID_ORGANIZATION_NAME),
        "country": _get_attr(x509.OID_COUNTRY_NAME),
        "state": _get_attr(x509.OID_STATE_OR_PROVINCE_NAME),
        "locality": _get_attr(x509.OID_LOCALITY_NAME),
        "valid_from": cert.not_valid_before.strftime("%Y-%m-%d"),
        "valid_to": cert.not_valid_after.strftime("%Y-%m-%d"),
    }


def get_cert_names_from_keyids(certs_dir, keyids):
    cert_names = []
    for keyid in keyids:
        try:
            name = extract_x509((Path(certs_dir) / keyid + ".pem").read_bytes())["name"]
            if not name:
                print("Cannot extract common name from x509, using key id instead.")
                cert_names.append(keyid)
            else:
                cert_names.append(name)
        except FileNotFoundError:
            print(f"Certificate does not exist ({keyid}).")
    return cert_names


def get_key_size(key_pem_path, decrypt_pwd=None):
    try:
        key = load_pem_public_key(Path(key_pem_path).read_bytes(), default_backend())
    except ValueError:
        key = load_pem_private_key(
            Path(key_pem_path).read_bytes(), decrypt_pwd, default_backend()
        )
    except Exception:
        return 0

    return key.key_size


def get_pin_for(name, confirm=True, repeat=True):
    pin = getpass(f"Enter PIN for {name}: ")
    if confirm:
        if pin != getpass(f"Confirm PIN for {name}: "):
            err_msg = "PINs don't match!"
            if repeat:
                print(err_msg)
                get_pin_for(name, confirm, repeat)
            else:
                raise PINMissmatchError(err_msg)
    return pin


def extract_json_objects_from_trusted_stdout(text, decoder=JSONDecoder()):
    """Find JSON objects in text, and yield the decoded JSON data

    Does not attempt to look for JSON arrays, text, or other JSON types outside
    of a parent JSON object.

    """
    pos = 0
    while True:
        match = text.find("{", pos)
        if match == -1:
            break
        try:
            result, index = decoder.raw_decode(text[match:])
            yield result
            pos = match + index
        except ValueError:
            pos = match + 1


def is_non_empty_directory(path: Path):
    if path.is_dir():
        return any(path.iterdir())
    return False


def is_run_from_python_executable() -> bool:
    """
    `sys frozen returns True if the Python interpreter is frozen using a tool like pyinstaller.
    """
    return getattr(sys, "frozen", False)


def read_input_dict(value):
    if value is None:
        return {}
    if not isinstance(value, dict):
        if Path(value).is_file():
            with open(value) as f:
                try:
                    value = json.loads(f.read())
                except json.decoder.JSONDecodeError:
                    print(f"\nWARNING: {value} is not a valid json!\n")
                    return {}

        else:
            try:
                value = json.loads(value)
            except json.decoder.JSONDecodeError:
                print(f"\nWARNING: {value} is not a valid json!\n")
                return {}

    return value


def run(*command, **kwargs):
    """Run a command and return its output. Call with `debug=True` to print to
    stdout.
    In order to get bytes, call this command with `raw=True` argument.
    """
    # Skip decoding

    raw = kwargs.pop("raw", False)
    data = kwargs.pop("input", None)
    timeout = kwargs.pop("timeout", None)

    if len(command) == 1 and isinstance(command[0], str):
        command = command[0].split()
    if taf.settings.LOG_COMMAND_OUTPUT:
        taf_logger.debug("About to run command {}", " ".join(command))

    def _format_word(word, **env):
        """To support words such as @{u} needed for git commands."""
        try:
            return word.format(env)
        except KeyError:
            return word

    command = [_format_word(word, **os.environ) for word in command]
    if timeout:
        command.append("--progress")
    try:
        options = dict(stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        if not raw:
            options.update(universal_newlines=True)
        if data is not None:
            options.update(input=data)

        options.update(kwargs)
        if not timeout:
            options.update(check=True)
            completed = subprocess.run(command, **options)
        else:
            options.update(text=True, bufsize=1)
            completed = run_with_timeout(command, options, timeout)

    except (subprocess.CalledProcessError, subprocess.TimeoutExpired) as err:
        if err.stdout:
            taf_logger.debug(err.stdout)
        if err.stderr:
            taf_logger.debug(err.stderr)
        taf_logger.debug(
            "Command {} returned non-zero exit status {} with output {}",
            " ".join(command),
            err.returncode,
            err.output,
        )
        raise err

    if completed.stdout:
        if taf.settings.LOG_COMMAND_OUTPUT:
            taf_logger.debug(completed.stdout)

    if completed.returncode != 0:
        return None

    return completed.stdout if raw else completed.stdout.rstrip()


def run_with_timeout(command, options, timeout=300):
    """Function to run a command with adaptive timeout and handle output."""
    buffer_size = 1024
    last_output_time = time.time()
    last_chunk = None
    with subprocess.Popen(command, **options) as proc:
        last_output_time = time.time()
        while True:
            # Read from stdout with a specific buffer size
            current_chunk = proc.stdout.read(buffer_size)
            if current_chunk:
                if last_chunk != current_chunk:
                    last_chunk = current_chunk
                    last_output_time = time.time()

            if time.time() - last_output_time > timeout:
                proc.kill()
                proc.wait()  # Ensure the process is cleaned up
                raise subprocess.CalledProcessError(
                    1,
                    proc.args,
                    output=None,
                    stderr=f"Command timed out after {timeout} seconds",
                )
            if proc.poll() is not None:
                break  # Process completed
            time.sleep(0.1)

        if proc.returncode == 0:
            return subprocess.CompletedProcess(
                proc.args, proc.returncode, stdout=last_chunk
            )
        else:
            raise subprocess.CalledProcessError(
                proc.returncode, proc.args, output="".join(last_chunk)
            )


def run_subprocess(command):
    try:
        result = subprocess.run(
            command, check=True, capture_output=True, text=True
        ).stdout
        return result
    except subprocess.CalledProcessError as e:
        taf_logger.error("An error occurred while executing {}: {}", command, e.output)
        raise e


def normalize_file_line_endings(file_path):
    with open(file_path, "rb") as open_file:
        content = open_file.read()
    replaced_content = normalize_line_endings(content)
    if replaced_content != content:
        with open(file_path, "wb") as open_file:
            open_file.write(replaced_content)


def normalize_line_endings(file_content):
    WINDOWS_LINE_ENDING = b"\r\n"
    UNIX_LINE_ENDING = b"\n"
    replaced_content = file_content.replace(
        WINDOWS_LINE_ENDING, UNIX_LINE_ENDING
    ).rstrip(UNIX_LINE_ENDING)
    return replaced_content


def on_rm_error(_func, path, _exc_info):
    """Used by when calling rmtree to ensure that readonly files and folders
    are deleted.
    """
    try:
        os.chmod(path, stat.S_IWRITE)
    except OSError as e:
        taf_logger.debug(f"File at path {path} not found, error trace - {e}")
        return
    try:
        os.unlink(path)
    except (OSError, PermissionError) as e:
        taf_logger.debug(
            "WARNING: Failed to clean up temporary update files: {}. This is a known issue when running TAF in a subprocess. You could consider upgrading taf to see if cleanup errors persist",
            e,
        )
        pass


def safely_save_json_to_disk(data, permanent_path):
    tfile = tempfile.NamedTemporaryFile(mode="w+t", delete=False)
    if data is not None:
        json.dump(data, tfile, indent=4)
    else:
        Path(permanent_path).write_text("")

    temp_file_path = tfile.name
    tfile.close()
    safely_move_file(temp_file_path, permanent_path, overwrite=True)


def safely_move_file(src, dst, overwrite=False):
    """Rename a file from ``src`` to ``dst``.

    *   Moves must be atomic.  ``shutil.move()`` is not atomic.
        Note that multiple threads may try to write to the cache at once,
        so atomicity is required to ensure the serving on one thread doesn't
        pick up a partially saved image from another thread.

    *   Moves must work across filesystems.  Often temp directories and the
        cache directories live on different filesystems.  ``os.rename()`` can
        throw errors if run across filesystems.

    So we try ``os.rename()``, but if we detect a cross-filesystem copy, we
    switch to ``shutil.move()`` with some wrappers to make it atomic.
    https://alexwlchan.net/2019/03/atomic-cross-filesystem-moves-in-python/
    MIT licensed.
    """
    src = str(src)
    dst = str(dst)
    if overwrite and os.path.isfile(dst):
        os.remove(dst)
    try:
        os.rename(src, dst)
    except OSError as err:
        if err.errno == errno.EXDEV:
            # Generate a unique ID, and copy `<src>` to the target directory
            # with a temporary name `<dst>.<ID>.tmp`.  Because we're copying
            # across a filesystem boundary, this initial copy may not be
            # atomic.  We intersperse a random UUID so if different processes
            # are copying into `<dst>`, they don't overlap in their tmp copies.
            copy_id = uuid.uuid4()
            tmp_dst = f"{dst}.{copy_id}.tmp"
            shutil.copyfile(src, tmp_dst)
            # Then do an atomic rename onto the new name, and clean up the
            # source image.
            os.rename(tmp_dst, dst)
            os.unlink(src)
        else:
            raise


def secret_yes_no_prompt(message: str) -> bool:
    """
    Ask user a yes/no question about loading more keys in a "hidden" manner.
    The user must type y or n explicitly. Any other input will re-prompt.
    """
    while True:
        response = click.prompt(
            message,
            hide_input=True,
            show_default=False,
            type=str,
        )
        response = response.strip().lower()

        if response in ("y", "n"):
            return response == "y"
        else:
            click.echo("Invalid input. Please type 'y' or 'n'.")


def to_tuf_datetime_format(start_date, interval):
    """Used to convert datetime to format used while writing metadata:
    e.g. "2020-05-29T21:59:34Z",
    """
    datetime_object = start_date + datetime.timedelta(interval)
    datetime_object = datetime_object.replace(microsecond=0)
    return datetime_object.isoformat() + "Z"


def resolve_keystore_path(
    keystore: Optional[str], roles_key_infos: Optional[str]
) -> str:
    if not keystore:
        return ""

    keystore_path = Path(keystore).expanduser().resolve()
    if roles_key_infos:
        roles_key_infos_path = Path(roles_key_infos)
        if roles_key_infos_path.is_file() and not keystore_path.is_absolute():
            keystore_path = (roles_key_infos_path.parent / keystore_path).resolve()

    return str(keystore_path)


def get_file_details(
    filepath: str,
    hash_algorithms: List[str] = ["sha256"],
    storage_backend: Optional[StorageBackendInterface] = None,
) -> Tuple[int, Dict[str, str]]:

    # Making sure that the format of 'filepath' is a path string.
    if not isinstance(filepath, str) or not filepath:
        raise ValueError("The filepath must be a non-empty string.")

    if not isinstance(hash_algorithms, list):
        raise ValueError("The hash_algorithms must be a list.")
    for algo in hash_algorithms:
        if algo not in ["sha256", "sha512"]:  # Add any other valid algorithms as needed
            raise ValueError(f"Invalid hash algorithm: {algo}")

    if storage_backend is None:
        storage_backend = FilesystemBackend()

    # Getting the file length
    if not os.path.isabs(filepath):
        raise ValueError("The 'filepath' must be an absolute path")

    # Check if the file exists and get its size
    if not os.path.isfile(filepath):
        raise FileNotFoundError(f"The file at '{filepath}' cannot be opened or found")

    file_length = os.path.getsize(filepath)

    # Getting the file hashes
    file_hashes = {}
    with storage_backend.get(filepath) as fileobj:
        for algorithm in hash_algorithms:
            digest_object = digest_fileobject(fileobj, algorithm)
            file_hashes.update({algorithm: digest_object.hexdigest()})
            fileobj.seek(
                0
            )  # Reset file object position after reading for the next hash

    return file_length, file_hashes


def ensure_pre_push_hook(auth_repo_path: Path) -> bool:
    hooks_dir = auth_repo_path / ".git" / "hooks"
    hooks_dir.mkdir(parents=True, exist_ok=True)
    pre_push_script = hooks_dir / "pre-push"
    resources_pre_push_script = (
        Path(__file__).parent / "resources" / "pre-push"
    ).resolve()

    # always copy the newest version of the pre-push hook

    shutil.copy(resources_pre_push_script, pre_push_script)
    try:
        if platform.system() != "Windows":
            # Unix-like systems
            pre_push_script.chmod(0o755)
    except Exception as e:
        taf_logger.error(f"Error setting executable permission: {e}")
        return False

    # Check if permissions were set correctly on Unix-like systems
    if platform.system() != "Windows" and not os.access(pre_push_script, os.X_OK):
        taf_logger.error(
            f"Failed to set pre-push git hook executable permission. Please set it manually for {pre_push_script}."
        )
        return False
    taf_logger.info("Pre-push hook updated successfully.")
    return True


class timed_run:
    """Decorator to let us capture the elapsed time and optionally print a timer and start/end
    messages around function calls"""

    def __init__(self, start_message=None, end_message="  completed in {} seconds"):
        self.start_message = start_message
        self.end_message = end_message
        self.start_time = None
        self.elapsed_time = None

    def start(self):
        self.start_time = time.time()
        if self.start_message is not None:
            # This message should be shown regardless of verbosity setting
            taf_logger.log("NOTICE", self.start_message)

    def end(self):
        self.elapsed_time = time.time() - self.start_time
        if self.end_message is not None:
            taf_logger.log("NOTICE", self.end_message.format(int(self.elapsed_time)))

    def __call__(self, orig_func=None):
        @wraps(orig_func)
        def wrapper_func(*args, **kwargs):
            self.start()
            result = orig_func(*args, **kwargs) if orig_func else None
            self.end()
            return result

        return wrapper_func


class TempPartition:
    def __init__(self, ref_path):
        self.ref_partition = self.get_partition_root(ref_path)
        temp_dir_partition = self.get_partition_root(Path(tempfile.gettempdir()))

        if temp_dir_partition == self.ref_partition:
            self.temp_dir = Path(tempfile.mkdtemp())
        else:
            taf_dir = self.ref_partition / ".taf"
            taf_dir.mkdir(exist_ok=True)
            self.temp_dir = tempfile.mkdtemp(dir=str(taf_dir))

    def get_partition_root(self, path):
        # Get the root directory of the partition containing the specified path
        while not os.path.ismount(path):
            path = path.parent
        return path

    def cleanup(self):
        # Remove the temporary directory and the ".taf" directory
        if Path(self.temp_dir).is_dir():
            shutil.rmtree(self.temp_dir, onerror=on_rm_error)

    def __enter__(self):
        return self.temp_dir, self.partition

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.cleanup()



import click
from pathlib import Path
from taf.api.metadata import update_metadata_expiration_date, check_expiration_dates, add_key_names
from taf.constants import DEFAULT_RSA_SIGNATURE_SCHEME
from taf.exceptions import TAFError
from taf.tools.cli import catch_cli_exception, common_repo_edit_options, find_repository
from taf.tools.repo import pin_managed
from taf.utils import ISO_DATE_PARAM_TYPE as ISO_DATE
from taf.log import taf_logger
import datetime


def add_key_names_command():
    @click.command(help="""If key names are not specified when creating repositories or adding a new role, "
        "this command can be used to specify names of keys given a name-public key mapping.""")
    @find_repository
    @common_repo_edit_options
    @click.option("--path", default=".", help="Authentication repository's location. If not specified, set to the current directory")
    @click.option("--commit-msg", default=None, help="Commit message")
    @pin_managed
    def adding_key_names(path, keys_description, keystore, pin_manager, no_commit, no_remote_check, prompt_for_keys, commit_msg):
        if not keys_description or Path(keys_description).is_file():
            taf_logger.error("Provide a path to an existing keys-description file")

        add_key_names(
            path=path,
            keys_description=Path(keys_description),
            keystore=keystore, pin_manager=pin_manager,
            commit=not no_commit,
            skip_remote_check=no_remote_check,
            prompt_for_keys=prompt_for_keys,
            commit_msg=commit_msg,
        )
    return adding_key_names


def check_expiration_dates_command():
    @click.command(help="""Check if the expiration dates of the metadata roles is still within an interval threshold.
        Expiration date is calculated by adding interval to start date. Interval is specified in days.
        The default value for interval in method is set to 30 days.
        Result contains metadata roles which have already expired and also roles which will expire (within the interval).
        Result is printed to the console and contains the following information:
        header:
            - start date
            - interval (in days)
        information:
            - role name
            - expiration date
        Example console output:
        Given a 30 day interval from today (2022-07-22):
            timestamp will expire on 2022-07-22
            snapshot will expire on 2022-07-28
            root will expire on 2022-08-19""")
    @find_repository
    @click.option("--path", default=".", help="Authentication repository's location. If not specified, set to the current directory")
    @click.option("--interval", default=30, type=int, help="Number of days added to the start date")
    @click.option("--start-date", default=datetime.datetime.now(), help="Date to which expiration interval is added", type=ISO_DATE)
    def checking_expiration_dates(path, interval, start_date):
        check_expiration_dates(path=path, interval=interval, start_date=start_date)
    return checking_expiration_dates


def update_expiration_dates_command():
    @click.command(help="""Update expiration date of the metadata file corresponding to the specified role.
        The new expiration date is calculated by adding interval to start date. The default
        value of the start date parameter is the current date, while default interval depends
        on the role and is:
            - 365 in case of root
            - 90  in case of targets
            - 7 in case of snapshot
            - 1 in case of timestamp and all other roles
        The updated metadata file is then signed. If the signing key should be loaded from
        the keystore file, it's necessary to specify its path when calling this command. If
        that is not the case, it will be needed to either enter the signing key directly or
        sign the file using a yubikey.

        If targets or other delegated role is updated, automatically sign snapshot and timestamp.""")
    @find_repository
    @catch_cli_exception(handle=TAFError)
    @common_repo_edit_options
    @click.option("--path", default=".", help="Authentication repository's location. If not specified, set to the current directory")
    @click.option("--role", multiple=True, help="A list of roles which expiration date should get updated")
    @click.option("--interval", default=None, type=int, help="Number of days added to the start date")
    @click.option("--scheme", default=DEFAULT_RSA_SIGNATURE_SCHEME, help="A signature scheme used for signing")
    @click.option("--start-date", default=datetime.datetime.now(), type=ISO_DATE, help="Date to which the interval is added")
    @click.option("--commit-msg", default=None, help="Commit message")
    @pin_managed
    def update_expiration_dates(path, role, interval, keystore, scheme, start_date, no_commit, prompt_for_keys, pin_manager, keys_description, no_remote_check, commit_msg):
        if not len(role):
            print("Specify at least one role")
            return
        update_metadata_expiration_date(
            path=path,
            pin_manager=pin_manager,
            roles=role,
            interval=interval,
            keystore=keystore,
            scheme=scheme,
            start_date=start_date,
            commit=not no_commit,
            prompt_for_keys=prompt_for_keys,
            keys_description=keys_description,
            skip_remote_check=no_remote_check,
            commit_msg=commit_msg,
        )
    return update_expiration_dates


def attach_to_group(group):
    group.add_command(add_key_names_command(), name="add-key-names")
    group.add_command(check_expiration_dates_command(), name='check-expiration-dates')
    group.add_command(update_expiration_dates_command(), name='update-expiration-dates')



from contextlib import contextmanager
import functools
from pathlib import Path
from typing import Callable, Dict, List, Optional, Union

from taf.api.utils._conf import find_keystore, read_keys_name_mapping_from_auth
from taf.auth_repo import AuthenticationRepository
from taf.constants import DEFAULT_RSA_SIGNATURE_SCHEME
from taf.exceptions import PushFailedError, TAFError
from taf.keys import load_signers
from taf.messages import git_commit_message
from taf.constants import METADATA_DIRECTORY_NAME
from taf.log import taf_logger
from taf.tuf.keys import _get_legacy_keyid


def key_management(
    roles: Optional[List[str]] = None,
    roles_fn: Optional[Callable[[Dict[str, str]], List[str]]] = None,
) -> Callable:
    """
    Decorator that:
    - Captures dynamic PIN arguments
    - Ensures necessary signers are loaded.
    - Cleans up PIN manager
    - Reads key name mappings if `keys_description` is provided.

    Parameters:
    - path (str): Path to the authentication repository.
    - roles (Optional[List[str]]): List of predefined roles to be loaded.
    - roles_fn (Optional[Callable[[Dict[str, str]], List[str]]]): Function that determines roles to be loaded.
    - keystore_path (Optional[str]): Path to the keystore.
    - keys_description (Optional[str]): Path to a key description file.

    Returns:
    - Callable: Decorated function with authentication management.
    """
    from taf.yubikey.yubikey_manager import manage_pins

    roles = roles or []

    def decorator(func: Callable) -> Callable:
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            if kwargs.get("auth_repo") is not None:
                auth_repo = kwargs.pop("auth_repo")
            else:
                path = kwargs.pop("path")
                auth_repo = AuthenticationRepository(path=path)

            with manage_pins() as pin_manager:
                auth_repo.pin_manager = pin_manager
                auth_repo = _setup_auth_repo_and_signers(
                    auth_repo, roles, roles_fn, **kwargs
                )
                kwargs["auth_repo"] = auth_repo
                return func(*args, **kwargs)

        return wrapper

    return decorator


@contextmanager
def key_management_context(
    roles: Optional[List[str]] = None,
    roles_fn: Optional[Callable[[Dict[str, str]], List[str]]] = None,
    **kwargs,
):
    """
    Context manager that does the same as the key_management decorator,
    returning the auth_repo instance instead of decorating a function.
    """
    from taf.yubikey.yubikey_manager import manage_pins

    roles = roles or []
    if kwargs.get("auth_repo") is not None:
        auth_repo = kwargs.pop("auth_repo")
        path = auth_repo.path
    else:
        path = kwargs.pop("path")
        auth_repo = AuthenticationRepository(path=path)

    with manage_pins() as pin_manager:
        auth_repo.pin_manager = pin_manager
        auth_repo = _setup_auth_repo_and_signers(auth_repo, roles, roles_fn, **kwargs)
        yield auth_repo


def _setup_auth_repo_and_signers(
    auth_repo: AuthenticationRepository,
    roles: Optional[List[str]],
    roles_fn: Optional[Callable[[Dict[str, str]], List[str]]] = None,
    **kwargs,
):
    """
    Extracts the common logic for initializing the auth_repo:
      - Get an existing auth_repo from kwargs or create a new one using the provided path.
      - Read key name mappings
      - Determine the complete list of roles, using roles and roles_fn.
      - Process a single key_pin.
      - Load signers for each role.

    Returns the auth_repo instance.
    """
    from taf.api.yubikey import get_yk_roles
    import taf.yubikey.yubikey as yk

    keystore_path = kwargs.get("keystore")

    keys_name_mappings = read_keys_name_mapping_from_auth(auth_repo)
    auth_repo.add_key_names(keys_name_mappings)

    all_roles = list(roles or [])
    if callable(roles_fn):
        roles_to_update = roles_fn(kwargs)
        if not isinstance(roles_to_update, list):
            taf_logger.error("roles_fn must return a list")
            raise TAFError("roles_fn must return a list")
        all_roles.extend(roles_to_update)
    all_roles = list(set(all_roles))
    if not all_roles:
        # Auto-detect roles via YubiKeys if none specified.
        roles_per_yks = get_yk_roles(auth_repo.path)
        for yk_roles in roles_per_yks.values():
            for role in yk_roles:
                if role not in all_roles:
                    all_roles.append(role)
    if not all_roles:
        taf_logger.error("No roles specified")
        raise TAFError("No roles specified")
    taf_logger.info(f"Loading keys of roles {', '.join(all_roles)}")

    key_id_pins = None

    if kwargs.get("key_pin") is not None:
        serial_nums = yk.get_serial_nums()
        if len(serial_nums) == 0:
            taf_logger.error("No Yubikeys found")
            raise TAFError(
                "Passed in a --key-pin but no YubiKeys inserted. Please insert a YubiKey and try again."
            )
        if len(serial_nums) > 1:
            taf_logger.error("Multiple Yubikeys found")
            raise TAFError(
                "Passed in a --key-pin but multiple YubiKeys inserted. Please insert only one YubiKey and try again."
            )
        serial_num = serial_nums[0]
        public_key = yk.get_piv_public_key_tuf(serial=serial_num)
        keyid = _get_legacy_keyid(public_key)
        key_id_pins = {keyid: kwargs.get("key_pin")}

    use_yubikeys_to_sign = key_id_pins is not None
    sorted_roles = sorted(all_roles, key=role_priority)
    auth_repo.pin_manager.auto_continue = True

    for role in sorted_roles:
        if not auth_repo.check_if_keys_loaded(role):
            keystore_signers, yubikey_signers = load_signers(
                auth_repo,
                role,
                keystore=keystore_path,
                key_id_pins=key_id_pins,
                use_yubikeys_to_sign=use_yubikeys_to_sign,
            )
            auth_repo.add_signers_to_cache({role: keystore_signers})
            auth_repo.add_signers_to_cache({role: yubikey_signers})

    return auth_repo


def _map_keynames_to_keyids(
    auth_repo: AuthenticationRepository, pin_args: Dict
) -> Dict:
    """
    Finds key ids based on key names and inserts
    key id and key pin, if specified via pin_args,
    into a dictionary.
    """
    key_id_pins: Dict = {}
    if not pin_args:
        return key_id_pins

    key_name_key_ids = auth_repo.get_key_ids_of_key_names(pin_args.keys())
    for key_name, key_id in key_name_key_ids.items():
        if key_name in pin_args:
            key_id_pins[key_id] = pin_args[key_name]

    if len(key_id_pins) != len(pin_args):
        all_roles = auth_repo.get_all_roles()
        for key_name, key_pin in pin_args.items():
            for role_name in all_roles:
                if role_name == key_name:
                    key_ids = auth_repo.get_key_ids_of_role(role_name)
                    if len(key_ids) == 1:
                        key_id_pins[key_ids[0]] = key_pin

    if len(key_id_pins) != len(pin_args):
        not_existing_names = [
            key_name for key_name in pin_args if key_name not in key_id_pins
        ]
        taf_logger.error(f"Keys {', '.join(not_existing_names)} not defined")
        raise TAFError(f"Keys {', '.join(not_existing_names)} not defined")
    return key_id_pins


@contextmanager
def transactional_execution(auth_repo):
    initial_commit = auth_repo.head_commit()
    try:
        yield
    except PushFailedError:
        pass
    except Exception:
        auth_repo.reset_to_commit(initial_commit, hard=True)
        raise


@contextmanager
def manage_repo_and_signers(
    auth_repo: AuthenticationRepository,
    roles: Optional[List[str]] = None,
    keystore: Optional[Union[str, Path]] = None,
    scheme: Optional[str] = DEFAULT_RSA_SIGNATURE_SCHEME,
    prompt_for_keys: Optional[bool] = False,
    paths_to_reset_on_error: Optional[List[Union[str, Path]]] = None,
    load_roles: Optional[bool] = True,
    load_parents: Optional[bool] = False,
    load_snapshot_and_timestamp: Optional[bool] = True,
    commit: Optional[bool] = True,
    push: Optional[bool] = True,
    commit_key: Optional[str] = None,
    commit_msg: Optional[str] = None,
    no_commit_warning: Optional[bool] = True,
):
    """
    A context manager that loads all signers and adds them to the specified authentication repository's
    signers cache. This allows for the execution of other methods without having to update the
    signers cache manually. Optionally, at the end, the context manager commits and pushes all changes made
    to the authentication repository and handles cleanup in case of an error.

    Arguments:
        auth_repo (AuthenticationRepository): Already instantiated authentication repository.
        roles (Optional[List[str]]): List of roles that are expected to be updated.
        keystore (Optional[Union[str, Path]]): Path to the keystore containing signing keys.
        scheme (Optional[str]): The signature scheme.
        prompt_for_keys (Optional[bool]): If True, prompts for keys if not found. Defaults to False.
        paths_to_reset_on_error (Optional[List[Union[str, Path]]]): Paths to reset if an error occurs.
        load_roles (Optional[bool]): If True, loads signing keys of the roles specified using the argument of the same name.
        load_parents (Optional[bool]): If true, loads sining keys of the specified roles' parents.
        load_snapshot_and_timestamp (Optional[bool]): If True, loads snapshot and timestamp signing keys.
        commit (Optional[bool]): If True, commits changes to the repository.
        push (Optional[bool]): If True, pushes changes to the remote repository.
        commit_key (Optional[str]): Commit key from `messages.py`
        commit_msg (Optional[str]): The message to use for commits.
        no_commit_warning (Optional[bool]): If True, suppresses warnings when not committing.
    """
    try:
        roles_to_load = set()
        if roles:
            unique_roles = set(roles)
            if load_roles:
                roles_to_load.update(unique_roles)
            if load_parents:
                roles_to_load.update(auth_repo.find_parents_of_roles(unique_roles))
        if load_snapshot_and_timestamp:
            roles_to_load.add("snapshot")
            roles_to_load.add("timestamp")
        if roles_to_load:
            if not keystore:
                keystore_path = find_keystore(auth_repo.path)
            else:
                keystore_path = Path(keystore)
            sorted_roles_to_load = sorted(roles_to_load, key=role_priority)
            for role in sorted_roles_to_load:
                if not auth_repo.check_if_keys_loaded(role):
                    keystore_signers, yubikey_signers = load_signers(
                        auth_repo,
                        role,
                        keystore=keystore_path,
                        scheme=scheme,
                        prompt_for_keys=prompt_for_keys,
                    )
                    auth_repo.add_signers_to_cache({role: keystore_signers})
                    auth_repo.add_signers_to_cache({role: yubikey_signers})
        yield
        if commit and auth_repo.something_to_commit():
            if not commit_msg and commit_key:
                commit_msg = git_commit_message(commit_key)
            auth_repo.commit_and_push(commit_msg=commit_msg, push=push)
        elif not no_commit_warning:
            taf_logger.log("NOTICE", "\nPlease commit manually\n")

    except PushFailedError as e:
        taf_logger.error(e)
        raise
    except Exception as e:
        taf_logger.error(f"An error occurred: {e}")
        if not paths_to_reset_on_error:
            paths_to_reset_on_error = [METADATA_DIRECTORY_NAME]
        elif METADATA_DIRECTORY_NAME not in paths_to_reset_on_error:
            paths_to_reset_on_error.append(METADATA_DIRECTORY_NAME)

        if auth_repo.is_git_repository and paths_to_reset_on_error:
            # restore metadata, leave targets as they might have been modified by the user
            # TODO flag for also resetting targets?
            # also update the CLI error handling]
            auth_repo.restore([str(path) for path in paths_to_reset_on_error])

        raise TAFError from e


def role_priority(role):
    """
    Return a numeric priority for the given role. This can be used as a sort order key.
    Example: sorted(roles, key=role_priority)

    The order is:
    0. root
    1. targets
    2. (delegated roles)
    3. snapshot
    4. timestamp

    Anything not explicitly listed is considered a "delegated role."
    """
    if role == "root":
        return 0
    elif role == "targets":
        return 1
    elif role == "snapshot":
        return 3
    elif role == "timestamp":
        return 4
    else:
        return 2



from __future__ import annotations
import datetime
import json
import itertools
import os
import re
import shutil
import uuid
import pygit2
import subprocess
import logging
from collections import OrderedDict
from functools import partial, reduce
from pathlib import Path

import requests  # type: ignore
from taf.models.types import Commitish
import taf.settings as settings
from taf.exceptions import (
    GitAccessDeniedException,
    NoRemoteError,
    NothingToCommitError,
    PushFailedError,
    TAFError,
    CloneRepoException,
    FetchException,
    InvalidRepositoryError,
    GitError,
    UpdateFailedError,
    PygitError,
)
from taf.log import NOTICE, taf_logger
from taf.utils import run
from typing import Callable, Dict, List, Optional, Tuple, Union
from .pygit import PyGitRepository

EMPTY_TREE = "4b825dc642cb6eb9a060e54bf8d69288fbee4904"


class GitRepository:
    def __init__(
        self,
        library_dir: Optional[Union[Path, str]] = None,
        name: Optional[str] = None,
        urls: Optional[List[str]] = None,
        custom: Optional[Dict] = None,
        default_branch: Optional[str] = None,
        allow_unsafe: Optional[bool] = False,
        path: Optional[Union[Path, str]] = None,
        alias: Optional[str] = None,
        *args,
        **kwargs,
    ):
        """
        Args:
          library_dir (Path or str): path to the library root. This is a directory which contains other repositories,
          whose full path is determined by appending their names to the library dir. A repository can be
          instantiated by specifying library dir and name, or by the full path.
          name (str): repository's name, which is appended to the library dir to form the full path.
          Must be in the namespace/name format. If library_dir is specified, name must be specified too.
          path (Path): repository's full filesystem path, which can be specified instead of name and library dir
          urls (list): repository's urls
          custom (dict): a dictionary containing other data
          default_branch (str): repository's default branch, automatically determined if not specified
          allow_unsafe: allow a git's security mechanism which prevents execution of git commands if
          the containing directory is owned by a different user to be ignored
          alias: Repository's alias, which will be used in logging statements to reference it
        """
        if isinstance(library_dir, str):
            library_dir = Path(library_dir)
        if isinstance(path, str):
            path = Path(path)

        if (library_dir, name).count(None) == 1:
            raise InvalidRepositoryError(
                "Both library_dir and name need to be specified"
            )
        if name is not None and library_dir is not None:
            self.name = self._validate_repo_name(name)
            self.path = self._validate_repo_path(library_dir, name, path)
            self.library_dir = library_dir.expanduser().resolve()
        elif path is None:
            raise InvalidRepositoryError(
                "Either specify library dir and name pair or path!"
            )
        else:
            # maintain support for repositories whose names are not of significance
            # in that case, only full path is specified (can happen if only using the GitRepository class)
            # without the rest of the framework in some cotext
            # name is still required for logging, so determine it based on the path
            # if the path points to a direcotry directly inside the root direcotry
            # set name to the name of that folder
            # otherwise, use the same format that is expected when name is specified
            if path.parent.parent != path.parent:
                self.name = f"{path.parent.name}/{path.name}"
                self.library_dir = path.parent.parent
            else:
                self.name = path.name
                self.library_dir = path.parent
            self.library_dir = self.library_dir.resolve()
            self.path = self._validate_repo_path(path)

        self.alias = alias
        self.urls = self._validate_urls([str(url) for url in urls]) if urls else None
        self.allow_unsafe = allow_unsafe
        self.custom = custom or {}
        if default_branch is None:
            default_branch = self._determine_default_branch()
        self.default_branch = default_branch

    _pygit = None

    @property
    def pygit(self):
        if self._pygit is None:
            if not self.is_git_repository:
                raise GitError(
                    self,
                    message=f"The path '{self.path.as_posix()}' is not a Git repository.",
                )
            try:
                self._pygit = PyGitRepository(self)
                if not self._pygit:
                    raise PygitError("PyGitRepository instance is None")

            except Exception as e:
                error_message = f"Failed to instantiate PyGitRepository: {e}"
                logging.error(error_message)
                raise PygitError(error_message)
        return self._pygit

    @classmethod
    def from_json_dict(cls, json_data: Dict):
        """Create a new instance based on data contained by the `json_data` dictionary,
        which can be create by calling `to_json_dict`
        """
        return cls(**json_data)

    def to_json_dict(self):
        """Returns a dictionary mapping all attributes to their values"""
        return {
            "library_dir": str(self.library_dir),
            "name": self.name,
            "urls": self.urls,
            "default_branch": self.default_branch,
            "custom": self.custom,
        }

    logging_functions = {
        logging.DEBUG: taf_logger.debug,
        logging.INFO: taf_logger.info,
        NOTICE: partial(taf_logger.log, "NOTICE"),
        logging.WARNING: taf_logger.warning,
        logging.ERROR: taf_logger.error,
        logging.CRITICAL: taf_logger.critical,
    }

    log_template = "{}{}"

    _remotes = None

    _is_bare_repo = None

    @property
    def remotes(self) -> List[str]:
        if self._remotes is None:
            repo = self.pygit_repo
            if repo is None:
                return []
            self._remotes = [remote.name for remote in repo.remotes]
        return self._remotes

    @property
    def is_detached_head(self) -> bool:
        repo = self.pygit_repo
        return repo.head_is_detached

    @property
    def is_git_repository(self) -> bool:
        """Check if the given path is the root of a Git repository."""
        # This is used when instantiating a PyGitRepository repo, so do not use
        # it here
        try:
            result = self._git("rev-parse --is-inside-work-tree", reraise_error=True)
            if result == "true":
                return True
            result = self._git("rev-parse --is-bare-repository", reraise_error=True)
            if result == "true":
                return True

        except GitError:
            return False
        return False

    @property
    def is_git_repository_root(self) -> bool:
        try:
            if not self.is_git_repository:
                return False
            repo = self.pygit_repo
            if repo is None:
                return False
            if self.is_bare_repository:
                return repo.is_bare and Path(repo.path).resolve() == self.path.resolve()
            else:
                git_path = self.path / ".git"
                return Path(repo.path).resolve() == git_path.resolve() and (
                    git_path.is_dir() or git_path.is_file()
                )
        except PygitError:
            return False

    @property
    def initial_commit(self) -> Commitish:
        return (
            Commitish.from_hash(
                self._git(
                    "rev-list --max-parents=0 HEAD", error_if_not_exists=False
                ).strip()
            )
            if self.is_git_repository
            else None
        )

    @property
    def log_prefix(self) -> str:
        if self.alias:
            return f"{self.alias}: "
        return f"Repo {self.name}: "

    @property
    def pygit_repo(self) -> pygit2.Repository:
        if self.pygit.repo is None:
            raise PygitError("Failed to instantiate PyGitRepository")
        return self.pygit.repo

    @property
    def is_bare_repository(self) -> bool:
        if self._is_bare_repo is None:
            if self.pygit_repo is not None:
                self._is_bare_repo = self.pygit_repo.is_bare
            else:
                raise GitError(
                    "Cannot determine if repository is a bare repository. Cannot instantiate pygit repository"
                )

        return self._is_bare_repo

    def _git(self, cmd, *args, **kwargs):
        """Call git commands in subprocess
        e.g.:
          self._git('checkout {}', branch_name)
        """
        log_error = kwargs.pop("log_error", False)
        log_error_msg = kwargs.pop("log_error_msg", "")
        reraise_error = kwargs.pop("reraise_error", False)
        log_success_msg = kwargs.pop("log_success_msg", "")
        error_if_not_exists = kwargs.pop("error_if_not_exists", True)

        if len(args):
            cmd = cmd.format(*args)
        if self.allow_unsafe:
            command = f"git -C {self.path} -c safe.directory={self.path} {cmd}"
        else:
            command = f"git -C {self.path} {cmd}"
        result = None
        if log_error or log_error_msg:
            try:
                result = run(command, **kwargs)
                if log_success_msg:
                    self._log_debug(log_success_msg)
            except subprocess.CalledProcessError as e:
                if error_if_not_exists and (
                    not self.path.is_dir() or not self.is_git_repository
                ):
                    log_error_msg = (
                        f"{self.path} does not exist or is not a git repository"
                    )
                    reraise_error = True
                if log_error_msg:
                    error = GitError(self, message=log_error_msg, error=e)
                    self._log_error(error.message)
                else:
                    error = GitError(self, command=command, error=e)
                    # not every git error indicates a problem
                    # if it does, we expect that either custom error message will be provided
                    # or that the error will be reraised
                    self._log_debug(error.message)
                if reraise_error:
                    raise error
        else:
            try:
                result = run(command, **kwargs)
            except subprocess.CalledProcessError as e:
                raise GitError(self, command=command, error=e)
            if log_success_msg:
                self._log_debug(log_success_msg)
        return result

    def _get_default_branch_from_local(self) -> str:
        """
        Get the default branch from the local repository.
        Try to get the default branch from `refs/remotes/origin/HEAD` first.

        In a repository where the refs/remotes/origin/HEAD does not exist, this error will trigger:
            fatal: ref refs/remotes/origin/HEAD is not a symbolic ref

        In those cases, get the default branch from `git remote show origin`.

        If the repository does not have a remote set, use `symbolic-ref` HEAD
        """
        try:
            branch = self._git(
                "symbolic-ref refs/remotes/origin/HEAD --short", reraise_error=True
            )
            return _local_branch_re.sub("", branch)
        except GitError as e:
            self._log_debug(f"Could not get remote HEAD at {self.path}: {e}")
            pass
        try:
            result = self._git("remote show origin", reraise_error=True)
            match = re.search(r"HEAD branch:(.*)", result)
            if match is not None:
                return match.group(1).strip()
        except (GitError, IndexError) as e:
            self._log_debug(
                f"Could not get HEAD branch with git remote show origin at {self.path}: {e}"
            )
            pass
        try:
            return self._git("symbolic-ref HEAD --short", reraise_error=True)
        except GitError as e:
            self._log_debug(f"Could not get HEAD branch at {self.path}: {e}")
            pass
        raise GitError(
            self,
            message="Could not determine default branch from local repository",
        )

    def _get_default_branch_from_remote(self, url: str) -> str:
        if not self.is_git_repository:
            raise GitError(
                self,
                message="Could not get default branch from remote. Not a git repository",
            )
        branch = self._git(
            f"ls-remote --symref {url} HEAD",
            log_error=True,
            log_error_msg="Unable to get default branch from remote",
            reraise_error=True,
        )
        branch = branch.split("\t", 1)[0]
        branch = branch.split()[-1]
        return _remote_branch_re.sub("", branch)

    def _log(self, log_func: Callable, message: str) -> None:
        log_func(self.log_template, self.log_prefix, message)

    def _log_debug(self, message: str) -> None:
        self._log(self.logging_functions[logging.DEBUG], message)

    def _log_info(self, message: str) -> None:
        self._log(self.logging_functions[logging.INFO], message)

    def _log_notice(self, message: str) -> None:
        self._log(self.logging_functions[NOTICE], message)

    def _log_warning(self, message: str) -> None:
        self._log(self.logging_functions[logging.WARNING], message)

    def _log_error(self, message: str) -> None:
        self._log(self.logging_functions[logging.ERROR], message)

    def _log_critical(self, message: str) -> None:
        self._log(self.logging_functions[logging.CRITICAL], message)

    def all_commits_on_branch(
        self, branch: Optional[str] = None, reverse: Optional[bool] = True
    ) -> List[Commitish]:
        """Returns a list of all commits on the specified branch.
        If branch is None, all commits on the currently checked out branch will be returned
        """
        repo = self.pygit_repo

        if branch:
            branch_obj = repo.branches.get(branch)
            if branch_obj is None:
                raise GitError(
                    self,
                    message=f"Error occurred while getting commits of branch {branch}. Branch does not exist",
                )
            latest_commit_id = branch_obj.target
        else:
            if self.head_commit() is None:
                raise GitError(
                    self,
                    message=f"Error occurred while getting commits of branch {branch}. No HEAD reference",
                )
            latest_commit_id = repo[repo.head.target].id

        sort = pygit2.GIT_SORT_REVERSE if reverse else pygit2.GIT_SORT_NONE
        commits = [
            Commitish.from_hash(commit.id.hex)
            for commit in repo.walk(latest_commit_id, sort)
        ]
        self._log_debug(
            f"found the following commits: {', '.join([commit.value for commit in commits])}"
        )
        return commits

    def all_commits_since_commit(
        self,
        since_commit: Optional[Commitish] = None,
        branch: Optional[str] = None,
        reverse: Optional[bool] = True,
    ) -> List[Commitish]:
        """Returns a list of all commits since the specified commit on the
        specified or currently checked out branch

        Raises:
            exceptions.GitError: An error occurred with provided commit SHA
        """

        if since_commit is None:
            try:
                return self.all_commits_on_branch(branch=branch, reverse=reverse)
            except GitError as e:
                self._log_warning(str(e))
                return []

        try:
            self.commit_exists(commit=since_commit)
        except GitError as e:
            self._log_warning(f"Commit {since_commit} not found in local repository.")
            raise e
        repo = self.pygit_repo

        if branch:
            branch_obj = repo.branches.get(branch)
            if branch_obj is None:
                return []
            latest_commit_id = branch_obj.target
        else:
            if self.head_commit() is None:
                return []
            latest_commit_id = repo[repo.head.target].id

        if repo.descendant_of(since_commit.hash, latest_commit_id):
            return []

        commits: List[Commitish] = []
        for commit in repo.walk(latest_commit_id):
            commit = Commitish.from_hash(commit.id.hex)
            if commit == since_commit:
                break
            commits.insert(0, commit)

        if not reverse:
            commits = commits[::-1]
        self._log_debug(
            f"found the following commits: {', '.join([commit.value for commit in commits])}"
        )
        return commits

    def add_remote(
        self, upstream_name: str, upstream_url: str, raise_error_if_exists=False
    ) -> None:
        try:
            if self.remote_exists(upstream_name):
                if raise_error_if_exists:
                    raise GitError(
                        self, message=f"Remote {upstream_name} already exists"
                    )
                return
            self._git("remote add {} {}", upstream_name, upstream_url)
            if self._remotes:
                self._remotes.append(upstream_name)
        except GitError as e:
            if "already exists" not in str(e):
                raise

    def branches(
        self, remote: bool = False, all: bool = False, strip_remote: bool = False
    ) -> List[str]:
        """Returns all branches."""
        repo = self.pygit_repo

        if all:
            branches = set(repo.branches)
        elif remote:
            branches = set(repo.branches.remote)
        else:
            branches = set(repo.branches.local)

        if strip_remote:
            remotes = self.remotes
            branches = set(
                [
                    reduce(lambda b, r: b.replace(f"{r}/", ""), remotes, branch)
                    for branch in branches
                ]
            )

        return list(branches)

    def branches_containing_commit(
        self,
        commit: Commitish,
        strip_remote: Optional[bool] = False,
        sort_key: Optional[Callable] = None,
    ) -> OrderedDict:
        """Finds all branches that contain the given commit"""
        repo = self.pygit_repo

        local_branches = remote_branches = []
        try:
            local_branches = list(repo.branches.local.with_commit(commit.hash))
        except pygit2.GitError:
            pass
        try:
            remote_branches = list(repo.branches.remote.with_commit(commit.hash))
        except pygit2.GitError:
            pass
        filtered_remote_branches = []
        if len(remote_branches):
            for branch in remote_branches:
                local_name = self.branch_local_name(branch)
                if (
                    local_name
                    and "HEAD" not in local_name
                    and local_name not in local_branches
                ):
                    if strip_remote:
                        branch = self.branch_local_name(branch)
                    filtered_remote_branches.append(branch)
        branches = {branch: False for branch in local_branches}
        branches.update({branch: True for branch in filtered_remote_branches})
        return OrderedDict(sorted(branches.items(), key=sort_key, reverse=True))

    def branch_exists(
        self, branch_name: str, include_remotes: Optional[bool] = True
    ) -> bool:
        """
        Checks if a branch with the given name exists.
        If include_remotes is set to True, this checks if
        a remote branch exists.
        """
        repo = self.pygit_repo

        branch = repo.branches.get(branch_name)
        # this git command should return the branch's name if it exists
        # empty string otherwise
        if branch is not None:
            return True
        if include_remotes:
            for remote in self.remotes:
                remote_branch = repo.branches.remote.get(f"{remote}/{branch_name}")
                if remote_branch is not None:
                    return remote_branch

                # finally, check remote branch
                if self.has_remote():
                    return branch_name in self._git(
                        f"ls-remote --heads origin {branch_name}",
                        log_error_msg=f"Repo {self.name}: could check if the branch exists in the remote repository",
                        reraise_error=True,
                    )

        return False

    def branch_off_commit(self, branch_name: str, commit: Commitish) -> None:
        """Create a new branch by branching off of the specified commit"""
        repo = self.pygit_repo

        try:
            pygit2_commit = repo[commit.hash]
            repo.branches.local.create(branch_name, pygit2_commit)
            branch = repo.lookup_branch(branch_name)
            ref = repo.lookup_reference(branch.name)
            repo.checkout(ref)
        except Exception as e:
            self._log_error(str(e))
            raise
        finally:
            self._log_info(
                f"Created a new branch {branch_name} from branching off of {commit}"
            )

    def branch_local_name(self, remote_branch_name: str) -> Optional[str]:
        """Strip remote from the given remote branch"""
        for remote in self.remotes:
            if remote_branch_name.startswith(remote + "/"):
                return remote_branch_name.split("/", 1)[1]
        return None

    def checkout_branch(
        self,
        branch_name: str,
        create: Optional[bool] = False,
        raise_anyway: bool = False,
    ) -> None:
        """Check out the specified branch. If it does not exists and
        the create parameter is set to True, create a new branch.
        If the branch does not exist and create is set to False,
        raise an exception."""
        repo = self.pygit_repo

        try:
            branch = repo.lookup_branch(branch_name)
            if branch is not None:
                ref = repo.lookup_reference(branch.name)
                repo.checkout(ref)
            else:
                self._git(
                    "checkout {}",
                    branch_name,
                    log_success_msg=f"Repo {self.name}: checked out branch {branch_name}",
                )
        except Exception as e:
            if raise_anyway:
                raise GitError(repo=self, message=str(e))
            # skip worktree errors
            if "the current HEAD of a linked repository" in str(e):
                return

            if create:
                self.create_and_checkout_branch(branch_name)
            else:
                self._log_error(f"could not checkout branch {branch_name}")
                raise (e)

    def check_if_clean_and_synced(self, branch: Optional[str] = None) -> bool:
        """
        Check if repository's worktree is clean and if the
        specified (or default) branch is synced with remote
        (if the repository has a remote set)
        """
        branch = branch or self.default_branch
        if self.something_to_commit():
            return False
        if not self.has_remote():
            return True
        if not self.synced_with_remote(branch):
            return False
        return True

    def checkout_paths(self, commit: Commitish, *args) -> None:
        repo = self.pygit_repo

        pygit_commit = repo.get(commit.hash)
        repo.checkout_tree(
            pygit_commit, paths=list(args), strategy=pygit2.GIT_CHECKOUT_FORCE
        )

    def checkout_orphan_branch(self, branch_name: str) -> None:
        """Creates orphan branch"""
        self._git(
            f"checkout --orphan {branch_name}", log_error=True, reraise_error=True
        )
        try:
            self._git("rm -rf .")
        except GitError:  # If repository is empty
            pass

    def check_files_exist(
        self, file_paths: List[str], commit: Optional[Commitish] = None
    ):
        """
        Check if file paths are known to git
        """
        existing_files: list = []
        non_existing: list = []
        repo = self.pygit_repo

        if commit is None:
            commit = self.head_commit()

        if commit is None:
            return existing_files, non_existing

        pygit_commit = repo[commit.hash]
        tree = pygit_commit.tree  # Get the tree of that commit

        for file_path in file_paths:
            try:
                # Check if the file exists in the tree
                tree[file_path]
                existing_files.append(file_path)
            except KeyError:
                non_existing.append(file_path)

        return existing_files, non_existing

    def clean(self):
        self._git("clean -fd")

    def cleanup(self):
        if self._pygit is not None:
            self._pygit.cleanup()
            self._pygit = None

    def clean_and_reset(self):
        """Cleans the untracked files and resets the HEAD to the latest commit."""
        try:
            self.clean()
            self.reset_to_head()
        except GitError as e:
            raise GitError(
                self, message=f"Failed to clean and reset the repository: {e}"
            )

    def clone(
        self, no_checkout: bool = False, bare: Optional[bool] = False, **kwargs
    ) -> None:
        self._log_info("cloning repository")

        self.path.mkdir(exist_ok=True, parents=True)
        if len(os.listdir(self.path)) != 0:
            raise GitError(
                repo=self,
                message=f"destination path {self.path} is not an empty directory.",
            )

        if self.urls is None:
            raise GitError(
                repo=self, message="cannot clone repository. No urls were specified"
            )
        params = []
        if bare:
            params.append("--bare")
        elif no_checkout:
            params.append("--no-checkout")

        for name, value in kwargs.items():
            if isinstance(value, bool):
                params.append(f"--{name}")
            else:
                params.append(f"--{name} {value}")

        joined_params = " ".join(params)

        cloned = False
        for url in self.urls:
            self._log_info(f"trying to clone from {url}")
            try:
                self._git(
                    "clone {} . {}",
                    url,
                    joined_params,
                    log_success_msg=f"successfully cloned from {url}",
                    reraise_error=True,
                    timeout=60,
                    error_if_not_exists=False,
                )
            except GitError as e:
                self._log_debug(f"could not clone from {url} due to {e}")
            else:
                self._log_info(f"successfully cloned from {url}")
                cloned = True
                break

        if not cloned:
            self.raise_git_access_error(CloneRepoException)

        if self.default_branch is None:
            self.default_branch = self._determine_default_branch()

    def clone_from_disk(
        self,
        local_path: Path,
        remote_url: Optional[str] = None,
        is_bare: bool = False,
        keep_remote=False,
        branches=None,
    ) -> None:
        self.path.mkdir(parents=True, exist_ok=True)
        pygit2.clone_repository(local_path, self.path, bare=is_bare)
        if not self.is_git_repository:
            raise GitError(
                self, message=f"Could not clone repository from local path {local_path}"
            )
        repo = self.pygit_repo

        if not keep_remote:
            self.remove_remote("origin")
            if remote_url is not None:
                self.add_remote("origin", remote_url)
                self.fetch()
                if repo is not None and branches:
                    local_branch_names = [
                        branch.split("/")[-1] for branch in repo.branches.local
                    ]
                    for branch in branches:
                        if branch in local_branch_names:
                            self.set_upstream(str(branch))

    # TODO test this
    def clone_or_pull(
        self,
        branches: Optional[List[str]] = None,
        only_fetch: Optional[bool] = False,
        **kwargs,
    ) -> Tuple[Commitish | None, Commitish | None]:
        """
        Clone or fetch the specified branch for the given repo.
        Return old and new HEAD.
        """
        try:
            old_head = self.head_commit()
        except GitError:
            # repo does not exist
            old_head = None

        if old_head is None:
            self._log_debug(f"cloning {self.name}")
            self._log_debug(f"old head sha is {old_head}")
            self.clone(**kwargs)
        else:
            if branches is None:
                default_branch = self.default_branch
                if default_branch is None:
                    raise FetchException(
                        "Cannot pull/clone repository. Branch not specified and default branch could not be determined"
                    )
                branches = [default_branch]
            self._log_debug(f"pulling branches {', '.join(branches)}")
            try:
                for branch in branches:
                    if only_fetch:
                        self._git("fetch", "origin", f"{branch}:{branch}")
                    else:
                        self._git("pull", "origin", branch)
                    self._log_info(f"successfully fetched branch {branch}")
            except GitError as e:
                if "fatal" in str(e):
                    raise FetchException(f"{self.path}: {str(e)}")
                pass

        new_head = self.head_commit()

        return old_head, new_head

    def create_and_checkout_branch(
        self, branch_name: str, raise_error_if_exists: Optional[bool] = True
    ) -> None:
        repo = self.pygit_repo

        try:
            branch = repo.lookup_branch(branch_name)
            if branch is not None and raise_error_if_exists:
                raise GitError(self, message=f"Branch {branch_name} already exists")
            try:
                commit = repo.revparse_single("HEAD")
                repo.branches.local.create(branch_name, commit)
                branch = repo.lookup_branch(branch_name)
                ref = repo.lookup_reference(branch.name)
                repo.checkout(ref)
            except KeyError:
                # this will be execute if there is no HEAD pointer
                flag = "-b" if raise_error_if_exists else "-B"
                self._git(
                    "checkout {} {}",
                    flag,
                    branch_name,
                    log_success_msg=f"created and checked out branch {branch_name}",
                    log_error=True,
                    reraise_error=True,
                )

        except Exception as e:
            self._log_error(str(e))
            raise
        finally:
            self._log_info(f"Created a new branch {branch_name}")

    def create_branch(
        self, branch_name: str, commit: Optional[Commitish] = None
    ) -> None:
        repo = self.pygit_repo

        try:
            if commit is not None:
                pygit_commit = repo[commit.hash]
            else:
                pygit_commit = repo.revparse_single("HEAD")
            repo.branches.local.create(branch_name, pygit_commit)
        except Exception as e:
            self._log_error(str(e))
            raise
        finally:
            self._log_info(f"Created a new branch {branch_name}")

    def checkout_commit(self, commit: Commitish) -> None:
        self._git(
            "checkout {}",
            commit.hash,
            log_success_msg=f"checked out commit {commit.hash}",
            log_error=True,
            reraise_error=True,
        )

    def branch_unpushed_commits(self, branch_name):
        repo = self.pygit_repo

        local_branch = repo.branches.get(branch_name)
        if local_branch is None:
            # local branch does not exist
            return False, []
        try:
            upstream_full_name = local_branch.upstream_name
        except KeyError:
            # no local branch => no unpushed local commit
            return False, []
        if not upstream_full_name:
            # no upstream branch - not pushed
            return True, []
        parts = upstream_full_name.split("/")
        upstream_name = "/".join(parts[2:])

        remote_branch = repo.branches.get(upstream_name)

        # Get the last common ancestor of local and remote branches
        merge_base = repo.merge_base(local_branch.target, remote_branch.target)

        # Check for commits in local branch since the merge base that are not in the remote branch
        unpushed_commits = []
        for commit in repo.walk(local_branch.target, pygit2.GIT_SORT_TOPOLOGICAL):
            if commit.id != merge_base and not repo.descendant_of(
                remote_branch.target, commit.id
            ):
                unpushed_commits.append(commit)
            else:
                break

        return bool(unpushed_commits), [
            Commitish.from_hash(commit.id) for commit in unpushed_commits
        ]

    def commit(
        self, message: str, paths_to_commit: Optional[List[str]] = None
    ) -> Commitish:
        if not paths_to_commit:
            self._git("add -A")
        else:
            self._git(f"add {' '.join(paths_to_commit)}")
        try:
            self._git("diff --cached --exit-code --shortstat", reraise_error=True)
        except GitError:
            try:
                run("git", "-C", str(self.path), "commit", "--quiet", "-m", message)
                return Commitish.from_hash(self._git("rev-parse HEAD"))
            except subprocess.CalledProcessError as e:
                raise GitError(
                    repo=self, message=f"could not commit changes due to:\n{e}", error=e
                )
        else:
            raise NothingToCommitError(repo=self, message="No changes to commit")

    def commit_empty(self, message: str) -> Commitish:
        run(
            "git",
            "-C",
            str(self.path),
            "commit",
            "--quiet",
            "--allow-empty",
            "-m",
            message,
        )
        return Commitish.from_hash(self._git("rev-parse HEAD"))

    def commit_exists(self, commit: Commitish) -> bool:
        try:
            return bool(self._git(f"rev-parse {commit.value}"))
        except GitError:
            return False

    def commits_on_branch_and_not_other(
        self, branch1: str, branch2: str
    ) -> List[Commitish]:
        """
        Meant to find commits belonging to a branch which branches off of
        a commit from another branch. For example, to find only commits
        on a speculative branch and not on the main branch.
        """
        merge_base = self.get_merge_base(branch1, branch2)
        commits = self.all_commits_since_commit(merge_base, branch1, reverse=False)

        return commits

    def commit_before_commit(self, commit: Commitish) -> Optional[Commitish]:
        repo = self.pygit_repo

        repo_commit_id = repo.get(commit.hash).id
        for comm in repo.walk(repo_commit_id):
            hex = comm.id.hex
            if hex != commit.hash:
                return Commitish.from_hash(hex)
        return None

    def create_local_branch_from_remote_tracking(self, branch, remote="origin"):
        repo = self.pygit_repo
        remote_branch_name = f"refs/remotes/{remote}/{branch}"
        remote_branch = repo.lookup_reference(remote_branch_name)
        if remote_branch is not None:
            local_branch = repo.lookup_branch(branch, pygit2.GIT_BRANCH_LOCAL)
            if local_branch is None:
                # Create a new local branch from the remote branch
                target_commit = repo[remote_branch.target]
                repo.create_branch(branch, target_commit)

    def delete_local_branch(self, branch_name: str) -> None:
        """Deletes local branch."""
        try:
            repo = self.pygit_repo

            repo.branches.delete(branch_name)
        except KeyError:
            raise GitError(
                repo=self,
                message=f"Could not delete branch {branch_name}. Branch does not exist",
            )
        except Exception:
            raise GitError(repo=self, message=f"Could not delete branch {branch_name}.")

    def delete_remote_tracking_branch(
        self, remote_branch_name: str, force: Optional[bool] = False
    ) -> None:
        flag = "-D" if force else "-d"
        self._git(f"branch {flag} -r {remote_branch_name}", log_error=True)

    def delete_remote_branch(
        self,
        branch_name: str,
        remote: Optional[str] = None,
        no_verify: Optional[bool] = False,
    ) -> None:
        """
        Delete remote branch.
        """
        if remote is None:
            remote = self.remotes[0]
        no_verify_flag = "--no-verify" if no_verify else ""
        self._git(
            f"push {remote} --delete {branch_name} {no_verify_flag}", log_error=True
        )

    def get_commit_date(self, commit: Commitish) -> str:
        """Returns commit date of the given commit"""
        repo = self.pygit_repo

        pygit_commit = repo.get(commit.hash)
        date = datetime.datetime.utcfromtimestamp(
            pygit_commit.commit_time + pygit_commit.commit_time_offset
        )
        formatted_date = date.strftime("%Y-%m-%d")
        return formatted_date

    def get_commit_message(self, commit: Commitish) -> str:
        """Returns commit message of the given commit"""
        repo = self.pygit_repo

        pygit_commit = repo.get(commit.hash)
        return pygit_commit.message

    def get_default_branch(self, url: Optional[str] = None) -> str:
        """Get the default branch of the repository. If url is provided, return the
        default branch from the remote. Otherwise, return the default
        branch from the local repository."""
        if url is not None:
            url = url.strip()
            return self._get_default_branch_from_remote(url)
        return self._get_default_branch_from_local()

    def get_json(
        self, commit: Commitish, path: str, raw: Optional[bool] = False
    ) -> Optional[Dict]:
        s = self.get_file(commit, path, raw=raw)
        if s and isinstance(s, str):
            return json.loads(s)
        return None

    def get_file(
        self,
        commit: Commitish,
        path: str,
        raw: Optional[bool] = False,
        with_id: Optional[bool] = False,
    ) -> Tuple[str, str] | str:
        path = Path(path).as_posix()
        try:
            git_id, content = self.pygit.get_file(commit, path, raw)
            if with_id:
                return git_id, content
            return content
        except TAFError as e:
            raise e
        except Exception:
            return self._git("show {}:{}", commit, path, raw=raw)

    def get_first_commit_on_branch(
        self, branch: Optional[str] = None
    ) -> Optional[Commitish]:
        branch = branch or self.default_branch
        first_commit = self._git(
            f"rev-list --max-parents=0 {branch}", error_if_not_exists=False
        )
        return Commitish.from_hash(first_commit.strip()) if first_commit else None

    def get_last_branch_by_committer_date(self) -> Optional[str]:
        """Find the latest branch based on committer date. Should only be used for
        testing purposes"""
        branches = self._git("branch --sort=committerdate").strip().split("\n")
        if not len(branches):
            return None
        return branches[-1]

    def get_remote_url(self) -> Optional[str]:
        try:
            return self._git("config --get remote.origin.url").strip()
        except GitError:
            return None

    def delete_branch(self, branch_name: str) -> None:
        self._git("branch -D {}", branch_name, log_error=True)

    def diff_between_revisions(
        self, revision1: Optional[str] = EMPTY_TREE, revision2: Optional[str] = "HEAD"
    ) -> str:
        return self._git("diff --name-status {} {}", revision1, revision2)

    def has_remote(self) -> bool:
        return len(self.remotes) > 0

    def head_commit(self) -> Optional[Commitish]:
        """Finds sha of the commit to which the current HEAD points"""
        repo = self.pygit_repo

        try:
            return Commitish.from_hash(repo.revparse_single("HEAD").id.hex)
        except Exception:
            return None

    def fetch(
        self,
        fetch_all: Optional[bool] = False,
        branch: Optional[str] = None,
        remote: Optional[str] = "origin",
    ) -> None:
        try:
            if fetch_all:
                self._git("fetch --all", log_error=True, reraise_error=True)
            else:
                if branch is None:
                    branch = ""
                self._git(
                    "fetch {} {}", remote, branch, log_error=True, reraise_error=True
                )
        except GitError:
            self.raise_git_access_error(operation="fetch")

    def fetch_from_disk(self, local_repo_path, branches):

        repo = self.pygit_repo
        temp_remote_name = f"temp_{uuid.uuid4().hex[:8]}"
        repo.remotes.create(temp_remote_name, local_repo_path)
        remote = repo.remotes[temp_remote_name]
        remote.fetch()

        for branch in branches:
            self.create_local_branch_from_remote_tracking(branch, temp_remote_name)

        repo.remotes.delete(temp_remote_name)

    def find_first_branch_matching_pattern(
        self,
        traverse_branch_name: str,
        pattern_func: Callable[[str], bool],
        include_remotes: bool = False,
        sort_key_func: Optional[Callable[[str], bool]] = None,
    ) -> Optional[str]:
        """
        Fetches changes from a local repository on disk.
        Temporarily adds it as a remote to the current repository
        and removes that remote after the operation is completed.
        """

        branch_tips = {}
        repo: pygit2.Repository = self.pygit_repo

        # Obtain the branch reference
        branch_ref = repo.lookup_branch(traverse_branch_name)
        # Ensure the branch exists
        if branch_ref is not None:
            # Get the target commit of the branch
            branch_target = branch_ref.target
        else:
            raise GitError(
                self, message=f"Branch {traverse_branch_name} does not exist"
            )

        branches = self.branches(all=include_remotes)
        all_branch_names = []

        for branch_name in branches:
            stripped_name = self._remove_remote_prefix(branch_name)
            if stripped_name in all_branch_names:
                continue
            if pattern_func(stripped_name):
                branch = repo.lookup_branch(branch_name)
                try:
                    branch_tips[stripped_name] = branch.peel().hex
                except Exception:
                    ref = repo.references[f"refs/remotes/{branch_name}"]
                    commit = ref.peel(pygit2.Commit)
                    branch_tips[stripped_name] = commit.hex
            all_branch_names.append(stripped_name)

        if sort_key_func is not None:
            all_branch_names = sorted(all_branch_names, key=sort_key_func, reverse=True)
        # Iterate over commits from newest to oldest
        if len(branch_tips):
            for commit in repo.walk(branch_target):
                for branch_name in all_branch_names:
                    tip_hex = branch_tips.get(branch_name)
                    if tip_hex is not None and (
                        commit.hex == tip_hex or repo.descendant_of(tip_hex, commit.hex)
                    ):
                        return branch_name
        return None

    def get_current_branch(self, full_name: Optional[bool] = False) -> str:
        """Return current branch."""
        repo = self.pygit_repo

        branch = repo.lookup_reference("HEAD").resolve()
        if full_name:
            return branch.name
        return branch.shorthand

    def get_last_remote_commit(
        self, url: Optional[str] = None, branch: Optional[str] = None
    ) -> Optional[Commitish]:
        """
        Fetch the last remote commit of the specified branch
        """
        branch = branch or self.default_branch
        if url is None:
            url = self.get_remote_url()
        if url is None:
            raise FetchException(
                "Could not fetch the last remote commit. URL not found"
            )
        last_commit = self._git(f"--no-pager ls-remote {url} {branch}", log_error=True)
        if last_commit:
            last_commit = last_commit.split("\t", 1)[0]
            # in some cases (e.g. upstream is defined the result might contain a warning line)
            return Commitish.from_hash(last_commit.split()[-1])
        return None

    def get_merge_base(self, branch1: str, branch2: str) -> Optional[Commitish]:
        """Finds the best common ancestor between two branches"""
        repo = self.pygit_repo

        commit1 = self.top_commit_of_branch(branch1)
        commit2 = self.top_commit_of_branch(branch2)
        if commit1 is None or commit2 is None:
            return None
        return Commitish.from_hash(repo.merge_base(commit1.hash, commit2.hash).hex)

    def get_tracking_branch(
        self, branch: Optional[str] = "", strip_remote: Optional[bool] = False
    ) -> Optional[str]:
        """Returns tracking branch name in format origin/branch-name or None if branch does not
        track remote branch.
        """
        try:
            tracking_branch = self._git(f"rev-parse --abbrev-ref {branch}@{{u}}")
            if strip_remote:
                tracking_branch = self.branch_local_name(tracking_branch)
            return tracking_branch
        except GitError:
            return None

    def init_repo(self, bare: Optional[bool] = False) -> None:
        if self.path.is_dir():
            self.path.mkdir(exist_ok=True, parents=True)
        flag = "--bare" if bare else ""
        self._git(f"init {flag}", error_if_not_exists=False)
        if self.urls is not None and len(self.urls):
            self._git("remote add origin {}", self.urls[0])

    def is_remote_branch(self, branch_name: str) -> bool:
        """
        Checks if the name of the specified branch is a valid remote branch name
        Does not check if the branch exists
        """
        for remote in self.remotes:
            if branch_name.startswith(remote + "/"):
                return True
        return False

    def list_files_at_revision(self, commit: Commitish, path: str = "") -> List[str]:
        posix_path = Path(path).as_posix()
        try:
            return self.pygit.list_files_at_revision(commit, posix_path)
        except TAFError as e:
            raise e
        except Exception:
            self._log_warning(
                "Perfomance regression: Could not list files with pygit2. Reverting to git subprocess"
            )
            return self._list_files_at_revision(commit, posix_path)

    def _list_files_at_revision(self, commit: Commitish, path: str) -> List[str]:
        if path is None:
            path = ""
        file_names = self._git("ls-tree -r --name-only {}", commit)
        list_of_files: List = []
        if not file_names:
            return list_of_files
        for file_in_repo in file_names.split("\n"):
            if not file_in_repo.startswith(path):
                continue
            file_in_repo = os.path.relpath(file_in_repo, path)
            list_of_files.append(file_in_repo)
        return list_of_files

    def list_changed_files_at_revision(self, commit: Commitish) -> List[str]:
        repo = self.pygit_repo

        commit1 = repo.get(commit.hash)
        commit_before_commit = self.commit_before_commit(commit)
        if commit_before_commit is not None:
            commit2 = repo.get(commit_before_commit.hash)

        diff = repo.diff(commit1, commit2)

        file_names = set()
        for patch in diff:
            delta = patch.delta
            file_names.add(delta.new_file.path)
            file_names.add(delta.old_file.path)
        return list(file_names)

    def list_commits(self, branch: Optional[str] = "") -> List[Commitish]:

        return [
            Commitish.from_hash(commit.hex)
            for commit in self.list_pygit_commits(branch)
        ]

    def list_pygit_commits(self, branch: Optional[str] = "") -> List[pygit2.Commit]:
        repo = self.pygit_repo

        if branch:
            branch_obj = repo.branches.get(branch)
            latest_commit_id = branch_obj.target
        else:
            latest_commit_id = repo[repo.head.target].id

        return [commit for commit in repo.walk(latest_commit_id, pygit2.GIT_SORT_NONE)]

    def list_n_commits(
        self,
        number: Optional[int] = 10,
        start_commit: Optional[Commitish] = None,
        branch: Optional[str] = None,
    ) -> List[Commitish]:
        """List the specified number of top commits of the current branch
        Optionally skip a number of commits from the top"""
        if number is None or number <= 0:
            return []
        repo = self.pygit_repo

        if start_commit is not None:
            start_commit_id = repo.get(start_commit.hash).id
        elif branch:
            branch_obj = repo.branches.get(branch)
            start_commit_id = branch_obj.target
        else:
            start_commit_id = repo[repo.head.target].id
        commits = itertools.islice(repo.walk(start_commit_id), number + 1)
        return [
            Commitish.from_hash(commit.hex)
            for commit in commits
            if commit.hex != start_commit_id
        ]

    def list_modified_files(
        self, path: Optional[str] = None, with_status: Optional[bool] = False
    ) -> List[Tuple]:
        # TODO
        # repo.diff("HEAD")
        # go over the returned diff object
        diff_command = "diff --name-status"
        if path is not None:
            diff_command = f"{diff_command} {path}"
        modified_files = self._git(diff_command).split("\n")
        files = []
        for modified_file in modified_files:
            # ignore warning lines
            if len(modified_file) and modified_file[0] in ["A", "M", "D"]:
                if with_status:
                    files.append(tuple(modified_file.split(maxsplit=1)))
                else:
                    files.append(modified_file.split(maxsplit=1)[1])
        return files

    def list_tags(self) -> List[str]:
        return self._git("tag -l").splitlines()

    def list_untracked_files(self, path: Optional[str] = None) -> List[str]:
        ls_command = "ls-files --others"
        if path is not None:
            ls_command = f"{ls_command} {path}"
        untracked_files = self._git(ls_command).split("\n")
        return [
            untracked_file for untracked_file in untracked_files if len(untracked_file)
        ]

    def merge_commit(
        self,
        commit: Commitish,
        target_branch: Optional[str] = None,
        fast_forward_only: Optional[bool] = False,
        check_if_merge_completed: Optional[bool] = False,
    ) -> bool:
        # Determine the branch to merge into, defaulting to the current branch if not provided
        branch = target_branch or self.get_current_branch()

        fast_forward_only_flag = "--ff-only" if fast_forward_only else ""

        self._git(f"merge {commit} {fast_forward_only_flag}", log_error=True)

        self.reset_to_commit(commit, branch, hard=True)
        if check_if_merge_completed:
            try:
                self._git("rev-parse -q --verify MERGE_HEAD")
                return False
            except GitError:
                pass
        return True

    def merge_branch(
        self, branch_name: str, allow_new_commit: Optional[bool] = False
    ) -> None:
        if allow_new_commit:
            repo = self.pygit_repo

            branch = repo.lookup_branch(branch_name)
            oid = branch.target

            for commit in repo.walk(oid):
                message = commit.message
                break

            repo.merge(oid)
            self.commit(message)
            repo.state_cleanup()
        else:
            self._git("merge {}", branch_name, log_error=True)

    def pull(self):
        """Pull current branch"""
        try:
            self._git("pull", log_error=True, reraise_error=True)
        except GitError:
            self.raise_git_access_error(operation="pull")

    def push(
        self,
        branch: Optional[str] = None,
        set_upstream: Optional[bool] = False,
        force: Optional[bool] = False,
        no_verify: Optional[bool] = False,
    ) -> bool:

        if not self.has_remote():
            self._log_warning("Could not push changes. No remotes configured")
            return False

        try:
            """Push all changes"""
            if branch is None:
                branch = self.get_current_branch()

            hook_path = Path(self.path) / ".git" / "hooks" / "pre-push"
            if hook_path.exists():
                self._log_notice("Validating and pushing...")

            upstream_flag = "-u" if set_upstream else ""
            force_flag = "-f" if force else ""
            no_verify_flag = "--no-verify" if no_verify else ""
            self._git(
                "push {} {} origin {} {}",
                upstream_flag,
                force_flag,
                branch,
                no_verify_flag,
                reraise_error=True,
            )
            self._log_notice("Successfully pushed to remote")
            return True
        except GitError as e:
            raise PushFailedError(self, message=f"Push operation failed: {e}")

    def raise_git_access_error(
        self, error_cls=GitAccessDeniedException, operation=None
    ):
        hosts = {extract_hostname(url) for url in self.urls}
        unknown_hosts = [host for host in hosts if not is_host_known(host)]
        if len(unknown_hosts):
            message = _no_hosts_error_format.format(hostname=",".join(unknown_hosts))
            raise error_cls(self, operation=operation, message=message)
        repo_exists = any(repository_exists(url) for url in self.urls)
        if repo_exists:
            uses_ssh = any(url.startswith("git@") for url in self.urls)
            if uses_ssh:
                raise error_cls(
                    self, operation=operation, message=_clone_or_pull_error_message
                )
            else:
                raise error_cls(
                    self,
                    operation=operation,
                    message=_clone_or_pull_error_message_no_ssh,
                )
        raise error_cls(self)

    def remove_remote(self, remote_name: str) -> None:
        try:
            self._git("remote remove {}", remote_name)
        except GitError as e:
            if "No such remote" not in str(e):
                self._git(f"remote rename {remote_name} local")
                self._log_warning(
                    f"Could not remove remote {remote_name}. It was renamed to 'local'. Remove it manually"
                )

    def remote_exists(self, remote_name):
        repo = self.pygit_repo
        for remote in repo.remotes:
            if remote.name == remote_name:
                return True
        return False

    def rename_branch(self, old_name: str, new_name: str) -> None:
        self._git("branch -m {} {}", old_name, new_name)

    def reset_num_of_commits(
        self, num_of_commits: int, hard: Optional[bool] = False
    ) -> None:
        flag = "--hard" if hard else "--soft"
        self._git(f"reset {flag} HEAD~{num_of_commits}")

    def reset_to_commit(
        self,
        commit: Commitish,
        branch: Optional[str] = None,
        hard: Optional[bool] = False,
        reset_remote_tracking=True,
    ) -> None:
        flag = "--hard" if hard else "--soft"

        if reset_remote_tracking:
            if branch is None:
                branch = self.get_current_branch()
            self.update_branch_refs(branch, commit)

        if hard:
            self._git(f"reset {flag} HEAD")

    def restore(self, file_paths: List[str]) -> None:
        if not file_paths:
            return
        file_paths = [str(Path(file_path).as_posix()) for file_path in file_paths]
        existing, non_existing = self.check_files_exist(file_paths)
        if existing:
            self._git(f"reset HEAD -- {' '.join(existing)}")
            self._git(f"checkout HEAD -- {' '.join(existing)}")
        for path in non_existing:
            file_path = Path(path)
            if file_path.is_file():
                file_path.unlink()
            elif file_path.is_dir():
                shutil.rmtree(file_path)

    def update_branch_refs(self, branch: str, commit: Commitish) -> None:
        # Update the local branch reference to the specific commit
        self._git(f"update-ref refs/heads/{branch} {commit}")
        # Update the remote-tracking branch
        self._git(f"update-ref refs/remotes/origin/{branch} {commit}", log_error=True)

    def update_ref_for_bare_repository(self, branch: str, commit: Commitish) -> None:
        """
        Update the reference of the branch to the given commit SHA in a bare repository.
        """
        try:
            self._git(f"update-ref refs/heads/{branch} {commit}")
        except GitError:
            raise UpdateFailedError(
                f"Could not update branch {branch} to commit {commit} in bare repository {self.name}."
            )

    def reset_remote_tracking_branch(self, branch_name) -> None:
        """
        Set top commit of origing/branch to the top comit of the local branch
        Used while testing the updater
        """
        commit = self.top_commit_of_branch(branch_name)
        self._git(f"update-ref refs/remotes/origin/{branch_name} {commit}")

    def reset_to_head(self) -> None:
        self._git("reset --hard HEAD")

    def safely_get_json(self, commit: Commitish, path: str) -> Optional[Dict]:
        try:
            return self.get_json(commit, path)
        except GitError:
            self._log_debug(f"{path} not available at revision {commit}")
        except json.decoder.JSONDecodeError:
            self._log_debug(f"{path} not a valid json at revision {commit}")
        return None

    def set_remote_url(self, new_url: str, remote: Optional[str] = "origin") -> None:
        self._git(f"remote set-url {remote} {new_url}")

    def set_upstream(self, branch_name: str) -> None:
        repo = self.pygit_repo
        try:
            repo.resolve_refish(f"origin/{branch_name}")
        except KeyError:
            return

        self._git("branch -u origin/{}", branch_name)

    def something_to_commit(self) -> bool:
        """Checks if there are any uncommitted changes"""
        if self.is_bare_repository:
            # For bare repositories, use `git diff` to check for uncommitted changes
            uncommitted_changes = self._git("diff --name-only --cached")
        else:
            # For non-bare repositories, use `git status`
            uncommitted_changes = self._git("status --porcelain")
        return bool(uncommitted_changes)

    def synced_with_remote(
        self, branch: Optional[str] = None, url: Optional[str] = None
    ) -> bool:
        """Checks if local branch is synced with its remote branch"""
        # check if the latest local commit matches
        # the latest remote commit on the specified branch

        if not self.has_remote():
            raise NoRemoteError(self)
        if url:
            urls = [url]
        elif self.urls:
            urls = self.urls
        else:
            remote_url = self.get_remote_url()
            if remote_url is None:
                raise GitError(
                    self,
                    message="URL not specified and could not be automatically determined. Cannot check if synced",
                )
            urls = [remote_url]

        branch_name = branch or self.default_branch
        if branch_name is None:
            raise GitError(
                self,
                message="Branch not specified and default branch could not be determined",
            )

        tracking_branch = self.get_tracking_branch(branch_name, strip_remote=True)
        if not tracking_branch:
            return False
        try:
            local_commit = self.top_commit_of_branch(branch_name)
        except GitError as e:
            if "unknown revision or path not in the working tree" not in str(e):
                raise e
            local_commit = None

        for url in urls:
            remote_commit = self.get_last_remote_commit(url, tracking_branch)
            if remote_commit is not None:
                break

        return local_commit == remote_commit

    def top_commit_of_branch(self, branch_name: str) -> Optional[Commitish]:
        repo = self.pygit_repo

        branch = repo.branches.get(branch_name)
        if branch is not None:
            return Commitish.from_hash(branch.target.hex)
        # a reference like HEAD
        try:
            return Commitish.from_hash(repo.revparse_single(branch_name).id.hex)
        except Exception:
            return None

    def update_local_branch(self, branch, remote_name="origin"):
        """
        Updates ref of a local branch of a bare repository where merging is not possible
        """
        repo = self.pygit_repo

        remote_branch_ref = f"refs/remotes/{remote_name}/{branch}"
        remote_branch_commit = repo.lookup_reference(remote_branch_ref).target

        # Update the local branch to point to the latest commit from the remote branch
        local_branch_ref = f"refs/heads/{branch}"
        repo.references.create(local_branch_ref, remote_branch_commit, force=True)

    def _determine_default_branch(self) -> Optional[str]:
        """Determine the default branch of the repository"""
        # try to get the default branch from the local repository
        errors = []
        try:
            return self.get_default_branch()
        except GitError as e:
            errors.append(e)
            pass

        # if the local repository does not exist or doesn't have a default branch, try to get the default branch from remote
        if self.urls:
            for url in self.urls:
                try:
                    return self.get_default_branch(url)
                except GitError as e:
                    errors.append(e)
                    pass

        self._log_debug(
            f"Cannot determine default branch with git -C at {self.path}: {errors}"
        )
        return None

    def top_commit_of_remote_branch(
        self, branch, remote="origin"
    ) -> Optional[Commitish]:
        """
        Fetches the top commit of the specified remote tracking branch.

        """
        remote_branch = f"{remote}/{branch}"
        if not self.branch_exists(remote_branch, include_remotes=True):
            raise GitError(
                self, message=f"Remote branch {remote_branch} does not exist"
            )
        return self.top_commit_of_branch(remote_branch)

    def _remove_remote_prefix(self, branch_name):
        for remote in self.remotes:
            prefix = f"{remote}/"
            if branch_name.startswith(prefix):
                return branch_name[len(prefix) :]
            return branch_name
        return branch_name

    def _validate_repo_name(self, name: str) -> str:
        """Ensure the repo name is not malicious"""
        match = _repo_name_re.match(name)
        if not match:
            taf_logger.error(f"Repository name {name} is not valid")
            raise InvalidRepositoryError(
                "Repository name must be in format namespace/repository "
                "and can only contain letters, numbers, underscores and "
                f'dashes, but got "{name}"'
            )
        return name

    def _validate_repo_path(
        self, library_dir: Path, name: Optional[str] = None, path: Optional[Path] = None
    ) -> Path:
        """
        validate repo path
        (since this is coming from potentially untrusted data)
        """
        # using OS to avoid resolve adding drive letter name on Windows
        repo_dir = os.path.join(str(library_dir), name or "")
        repo_dir = os.path.normpath(repo_dir)
        if not repo_dir.startswith(str(Path(library_dir))):
            taf_logger.error(f"Repository path/name {library_dir}/{name} is not valid")
            raise InvalidRepositoryError(
                f"Repository path/name {library_dir}/{name} is not valid"
            )
        repo_dir_path = Path(repo_dir).expanduser().resolve()
        if path is not None and path != repo_dir_path:
            raise InvalidRepositoryError(
                "Both library dir and name pair and path specified and are not equal. Omit the path."
            )
        return repo_dir_path

    def _validate_url(self, url: str) -> None:
        """ensure valid URL"""
        for _url_re in [_http_fttp_url, _ssh_url]:
            match = _url_re.match(url)
            if match:
                return
        self._log_error(f"URL ({url}) is not valid")
        raise InvalidRepositoryError(
            f'Repository URL must be a valid URL, but got "{url}".'
        )

    def _validate_urls(self, urls: Optional[List[str]]) -> Optional[List[str]]:
        def _find_url(path, url):
            try:
                if (path / url).resolve().is_dir():
                    return str((path / url).resolve())
            except OSError:
                pass
            if os.path.isabs(url):
                return url
            return str((path).resolve())

        if urls is not None:
            if settings.update_from_filesystem is False:
                for url in urls:
                    self._validate_url(url)
            else:
                # resolve paths and deduplicate
                urls = sorted((_find_url(self.path, url) for url in urls), reverse=True)
        return urls


_clone_or_pull_error_message = (
    "The remote repository exists, so this is probably due to the lack of privileges or SSH key issues. "
    "Here are some steps you can take to resolve common issues:\n\n"
    "1. Check Repository Access:\n"
    "   - If the repository is private, verify that your account has been granted access.\n\n"
    "2. Check SSH Key Configuration:\n"
    "   - Ensure your SSH key is correctly added to your Git hosting account (GitHub, GitLab, Bitbucket, etc.).\n"
    "   - Verify that your SSH key does not require a passphrase, or use an SSH agent to manage the passphrase.\n\n"
    "3. Using an SSH Agent:\n"
    "   - If your SSH key is passphrase-protected, start the ssh-agent in the background:\n"
    "       eval $(ssh-agent -s)\n"
    "   - Add your SSH key to the ssh-agent:\n"
    "       ssh-add /path/to/your/private/key\n"
    "4. If Problems Persist:\n"
    "   - If you are using Windows and encounter issues, consider running these commands in Bash or another Unix-like shell available on Windows, such as Git Bash or WSL.\n"
    "   - Try using a different SSH key that does not require a passphrase.\n"
)


_clone_or_pull_error_message_no_ssh = (
    "The remote repository exists, so this issue is probably due to lack of privileges.\n"
    "Verify that you have access to the repository if it is private, and verify your HTTPS configuration.\n"
    "Consider switching to SSH for potentially enhanced security and easier handling of credentials."
)


_no_hosts_error_format = (
    "Error: The host(s) '{hostname}' not recognized."
    "To add a host to your known_hosts file, manually connect via SSH using the command:\n"
    "ssh -T git@{hostname}\n"
    "Accept the host key to proceed.\n"
)

_remote_branch_re = re.compile(r"^(refs/heads/)")

_local_branch_re = re.compile(r"^(origin/)")

_repo_name_re = re.compile(r"^\w[\w_-]*/\w[\w_-]*$")

_http_fttp_url = re.compile(
    r"^(?:http|ftp)s?://"  # http:// or https://
    # domain...
    r"(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|"
    r"localhost|"  # localhost...
    r"\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})"  # ...or ip
    r"(?::\d+)?"  # optional port
    r"(?:/?|[/?]\S+)$",
    re.IGNORECASE,
)

_ssh_url = re.compile(
    r"((git|ssh|http(s)?)|(git@[\w\.]+))(:(//)?)([\w\.@\:/\-~]+)(\.git)?(/)?"
)


def extract_hostname(url):
    """Extract the hostname from a Git URL, which can be either SSH or HTTP(S)."""
    if url.startswith("git@"):
        # Handle SSH URL
        match = re.match(r"git@(.*?):", url)
    else:
        # Handle HTTP(S) URL
        match = re.match(r"https?://([^/]+)/", url)

    if match:
        return match.group(1)
    else:
        return None


def is_host_known(hostname):
    """Check if the given hostname is in the known_hosts file."""
    known_hosts_path = os.path.expanduser("~/.ssh/known_hosts")
    try:
        with open(known_hosts_path, "r") as file:
            for line in file:
                if line.startswith(hostname) or hostname in line.split():
                    return True
        return False
    except FileNotFoundError:
        print("Known hosts file not found.")
        return False
    except Exception as e:
        print(f"An error occurred: {str(e)}")
        return False


def repository_exists(url):
    """Check if the repository URL exists by making a HEAD request to the URL."""
    # Convert SSH URL to HTTP URL
    if url.startswith("git@"):
        # General pattern: git@HOST:USER/REPO
        http_url = re.sub(r"git@(.*?):(.*?)/(.*)\.git", r"https://\1/\2/\3", url)
    elif url.startswith("git://"):
        http_url = url.replace("git://", "https://").replace(".git", "")
    else:
        http_url = url

    try:
        response = requests.head(http_url)
        return response.status_code == 200
    except requests.RequestException as e:
        print(f"Error checking repository URL: {e}")
        return False



import re
from pathlib import Path
import attrs
from typing import Any, List, Optional


def integer_validator(
    instance: Any,
    attribute: "attrs.Attribute[int]",
    value: Any,
) -> bool:
    """Validates that integer is a number >=1"""
    if not isinstance(value, int) or not (value >= 1):
        name = attribute.name if hasattr(attribute, "name") else str(attribute)
        raise ValueError(
            f"{instance.__class__.__qualname__}.{name} should be a number >= 1, but was {value}."
        )
    return True


def public_key_validator(
    instance: Any,
    attribute: "attrs.Attribute[str]",
    value: Any,
) -> bool:
    if value and not value.startswith("-----BEGIN PUBLIC KEY-----"):
        raise ValueError("Public key must start with '-----BEGIN PUBLIC KEY-----'")
    return True


def role_number_validator(
    instance: Any,
    attribute: "attrs.Attribute[int]",
    value: int,
) -> bool:

    prefix = f"{instance.name if hasattr(instance, 'name') else 'delegated role'} definition error:"
    if not isinstance(value, int) or value < 1:
        raise ValueError(f"{prefix} number must be an integer greater than 1")
    if value < instance.threshold:
        raise ValueError(
            f"{prefix} number of signing keys ({value}) is lower than the threshold ({instance.threshold})"
        )
    if (
        value is not None
        and instance.yubikeys is not None
        and len(instance.yubikeys)
        and value != len(instance.yubikeys)
    ):
        raise ValueError(
            f"{prefix} number of signing keys ({value}) is not the same as the number of specified yubikeys. "
            "Omit the number property if specifying yubikeys or make sure these values match"
        )
    return True


def role_paths_validator(
    instance: Any,
    attribute: "attrs.Attribute[List[str]]",
    value: Optional[List[str]],
) -> bool:
    if value is None:
        return False
    word_pattern = r"^[A-Za-z0-9_.-]*$"

    for path in value:
        words = path.split("/")

        for word in words:
            if not re.match(word_pattern, word) and word != "*":
                raise ValueError(
                    f"{path} is not a valid delegated path. Delegated paths are valid directory names or * separated by /"
                )

    return True


def filepath_validator(
    instance: Any,
    attribute: "attrs.Attribute[List[str]]",
    value: str,
) -> bool:
    if value is not None and not Path(value).resolve().exists():
        raise ValueError(f"{value} does not exist")
    return True


def optional_type_validator(type):
    return attrs.validators.optional(attrs.validators.instance_of(type))



import logging

# Set a directory that should be used for all temporary files. If this
# is None, then the system default will be used. The system default
# will also be used if a directory path set here is invalid or
# unusable.
temporary_directory = None

conf_directory_root = None

update_from_filesystem = False

validate_repo_name = True

validation_repo_path: dict = {}

default_branch = None

# Strict mode enabled/disabled. If strict is enabled, any warnings
# should raise TAF errors
strict = False

# Allows usage of specified commit as last validated commit
# Useful when validating a local repository
overwrite_last_validated_commit = False

last_validated_commit: dict = {}

# determines if script files will be loaded from disk
development_mode = False

# determines if lifecycle handler scripts will be run
run_scripts = False

# The 'log.py' module manages TUF's logging system.  Users have the option to
# enable/disable logging to a file via 'ENABLE_FILE_LOGGING', or
# tuf.log.enable_file_logging() and tuf.log.disable_file_logging().


# whether to log to the console, which does not mean that file logging cannot be enabled as well
ENABLE_CONSOLE_LOGGING = True

# whether to log to a file, which does not mean that console logging cannot be enabled as well
ENABLE_FILE_LOGGING = True

# should errors be logged to a separate file if logging to file is enabled
SEPARATE_ERRORS = True

CONSOLE_LOGGING_LEVEL = logging.INFO

FILE_LOGGING_LEVEL = logging.INFO

ERROR_LOGGING_LEVEL = logging.WARNING

DEBUG_LOGGING_LEVEL = logging.DEBUG

# Location of the log files. It can be specified by setting LOGS_LOCATION
# and by setting an environment variable called TAF_LOG.
# If this location is not specified logs will be placed ~/.taf
LOGS_LOCATION = None

LOG_FILENAME = "taf.log"

ERROR_LOG_FILENAME = "taf.err"

DEBUG_LOG_FILENAME = "taf-debug.log"

LOG_COMMAND_OUTPUT = False

VERBOSITY = 0



import os
import sys
import logging
from typing import Dict
from pathlib import Path

from loguru import logger as taf_logger
import taf.settings as settings

_CONSOLE_FORMAT_STRING = "\n{message}\n"
_FILE_FORMAT_STRING = "[{time}] [{level}] [{module}:{function}@{line}]\n{message}\n"

console_loggers: Dict = {}
file_loggers: Dict = {}

NOTICE = 25
taf_logger.level("NOTICE", no=NOTICE, color="<yellow>", icon="!")

VERBOSITY_LEVELS = {
    0: "NOTICE",  # Default
    1: "INFO",  # -v
    2: "DEBUG",  # -vv
    3: "TRACE",  # -vvv
}


def formatter(record):
    if record["level"].no == NOTICE:
        return f"<white>{_CONSOLE_FORMAT_STRING}</white>"
    elif record["level"].no == logging.WARNING:
        return f"<yellow>{_CONSOLE_FORMAT_STRING}</yellow>"
    elif record["level"].no == logging.INFO:
        return f"<blue>{_CONSOLE_FORMAT_STRING}</blue>"
    elif record["level"].no == logging.DEBUG:
        return f"<magenta>{_CONSOLE_FORMAT_STRING}</magenta>"
    elif record["level"].no == logging.ERROR:
        return f"<red>{_CONSOLE_FORMAT_STRING}</red>"
    else:
        return _CONSOLE_FORMAT_STRING


def disable_console_logging():
    try:
        taf_logger.remove(console_loggers["log"])
    except ValueError:
        # will be raised if this is called twice
        pass


def disable_file_logging():
    try:
        taf_logger.remove(file_loggers["log"])
    except ValueError:
        # will be raised if this is called twice
        pass


def _get_log_location():
    location = settings.LOGS_LOCATION or os.environ.get("TAF_LOG")
    if location is None:
        location = Path.home() / ".taf"
        location.mkdir(exist_ok=True)
    else:
        location = Path(location)
    return location


def initialize_logger_handlers():
    taf_logger.remove()
    if settings.ENABLE_CONSOLE_LOGGING:
        console_loggers["log"] = taf_logger.add(
            sys.stdout, format=formatter, level=VERBOSITY_LEVELS[settings.VERBOSITY]
        )

    if settings.ENABLE_FILE_LOGGING:
        log_location = _get_log_location()
        log_path = str(log_location / settings.LOG_FILENAME)
        file_loggers["log"] = taf_logger.add(
            log_path,
            format=_FILE_FORMAT_STRING,
            level=settings.FILE_LOGGING_LEVEL,
            rotation="30 MB",
            retention=5,
            compression="zip",
        )

        if settings.SEPARATE_ERRORS:
            error_log_path = str(log_location / settings.ERROR_LOG_FILENAME)
            file_loggers["error"] = taf_logger.add(
                error_log_path,
                format=_FILE_FORMAT_STRING,
                level=settings.ERROR_LOGGING_LEVEL,
                rotation="30 MB",
                retention=5,
                compression="zip",
            )

        debug_log_path = str(log_location / settings.DEBUG_LOG_FILENAME)
        file_loggers["debug"] = taf_logger.add(
            debug_log_path,
            format=_FILE_FORMAT_STRING,
            level=settings.DEBUG_LOGGING_LEVEL,
            rotation="30 MB",
            retention=5,
            compression="zip",
        )


initialize_logger_handlers()



import json
from pathlib import Path
import click
from taf.api.roles import (
    add_roles as add_multiple_roles,
    list_keys_of_role,
    add_signing_key,
    remove_role,
    revoke_signing_key,
    remove_paths,
    rotate_signing_key,
)
from taf.constants import DEFAULT_RSA_SIGNATURE_SCHEME
from taf.exceptions import TAFError
from taf.auth_repo import AuthenticationRepository
from taf.log import taf_logger
from taf.tools.cli import catch_cli_exception, common_repo_edit_options, find_repository

from taf.api.roles import add_role_paths
from taf.tools.repo import pin_managed


def add_roles_command():
    @click.command(help="""
    Add new delegated target roles. Allows optional specification of each role's properties through a JSON configuration file.

    Configuration file (JSON) can specify:
    - 'parent_role' (string): The parent role under which the new role will be delegated. Default is 'targets'.
    - 'delegated_path' (array of strings): Paths to be delegated to the new role. Must specify at least one if using a config file.
    - 'keys_number' (integer): Number of signing keys. Default is 1.
    - 'threshold' (integer): Number of keys required to sign. Default is 1.
    - 'yubikey' (boolean): Whether to use a YubiKey for signing. Default is false.
    - 'scheme' (string): Signature scheme, e.g., 'rsa-pkcs1v15-sha256'. Default is 'rsa-pkcs1v15-sha256'.

    The structure of the configuration file is the very similar tohe structure of the one use when creating the repository,
    but does not have to contain all roles, just the ones that should be added. If roles that already
    exist are also defined, they will be skipped. If the role's parent is not the main targets role,
    it's necessary to specify it using the "parent_role" option

    Example JSON structure:
    {
    "yubikeys": {
        "user1": {
        "public": "<public_key>",
        "scheme": "rsa-pkcs1v15-sha256",
        "present": false
        },
        "userYK": {
        "scheme": "rsa-pkcs1v15-sha256"
        }
    },
    "roles": {
        "<role name>": {
        "parent_role": "targets",
        "paths": [
            "</delegated_path_inside_targets1>",
            "</delegated_path_inside_targets2>"
        ],
        "number": 2,
        "threshold": 1,
        "yubikey": true,
        "scheme": "rsa-pkcs1v15-sha256",
        "yubikeys": [
            "user1",
            "userYK"
        ]
        }
    }
    }
    """)
    @find_repository
    @catch_cli_exception(handle=TAFError)
    @common_repo_edit_options
    @click.option("--config-file", type=click.Path(exists=True), help="Path to the JSON configuration file.")
    @click.option("--path", default=".", help="Authentication repository's location. If not specified, set to the current directory")
    @click.option("--scheme", default=DEFAULT_RSA_SIGNATURE_SCHEME, help="A signature scheme used for signing")
    @pin_managed
    def add_roles(config_file, path, scheme, keystore, no_commit, prompt_for_keys, pin_manager, keys_description, no_remote_check):
        add_multiple_roles(
            path=path,
            pin_manager=pin_manager,
            keystore=keystore,
            roles_key_infos=config_file,
            scheme=scheme,
            prompt_for_keys=prompt_for_keys,
            commit=not no_commit,
            keys_description=keys_description,
            skip_remote_check=no_remote_check,
        )
    return add_roles


def export_roles_description_command():
    @click.command(help="""
        Export roles-description.json file based on the
        """)
    @find_repository
    @catch_cli_exception(handle=TAFError)
    @click.option("--path", default=".", help="Authentication repository's location. If not specified, set to the current directory")
    @click.option("--output", default=None, help="Output file path")
    @click.option("--keystore", default=None, help="Location of the keystore files")
    @pin_managed
    def export_roles_description(path, output, keystore, pin_manager):
        auth_repo = AuthenticationRepository(path=path, pin_manager=pin_manager)
        roles_description = auth_repo.generate_roles_description()
        if keystore:
            roles_description["keystore"] = keystore
        if not output:
            taf_logger.log("NOTICE", json.dumps(roles_description, indent=True))
        else:
            output = Path(output)
            output.parent.mkdir(exist_ok=True, parents=True)
            output.write_text(json.dumps(roles_description, indent=True))

    return export_roles_description


def add_role_paths_command():
    @click.command(help="Add a new delegated target role, specifying which paths are delegated to the new role. Its parent role, number of signing keys and signatures threshold can also be defined. Update and sign all metadata files and commit.")
    @find_repository
    @catch_cli_exception(handle=TAFError)
    @click.argument("role")
    @common_repo_edit_options
    @click.option("--path", default=".", help="Authentication repository's location. If not specified, set to the current directory")
    @click.option("--delegated-path", multiple=True, help="Paths associated with the delegated role")
    @pin_managed
    def adding_role_paths(role, path, delegated_path, keystore, no_commit, prompt_for_keys, pin_manager, keys_description, no_remote_check):
        if not delegated_path:
            print("Specify at least one path")
            return

        add_role_paths(
            path=path,
            paths=delegated_path,
            pin_manager=pin_manager,
            delegated_role=role,
            keystore=keystore,
            commit=not no_commit,
            prompt_for_keys=prompt_for_keys,
            push=not no_commit,
            keys_description=keys_description,
            skip_remote_check=no_remote_check,
        )
    return adding_role_paths


# commenting out this command since its execution leads to an invalid state
# this is a TUF bug (or better said, caused by using a newer version of the updater and old repository_tool)
# it will be addressed when we transition to metadata API
def remove_role_command():
    @click.command(help="""Remove a delegated target role, and, optionally, its targets (depending on the remove-targets parameter).
        If targets should also be deleted, target files are remove and their corresponding entires are removed
        from repositoires.json. If targets should not get removed, the target files are signed using the
        removed role's parent role
        """)
    @find_repository
    @catch_cli_exception(handle=TAFError)
    @click.argument("role")
    @common_repo_edit_options
    @click.option("--path", default=".", help="Authentication repository's location. If not specified, set to the current directory")
    @click.option("--scheme", default=DEFAULT_RSA_SIGNATURE_SCHEME, help="A signature scheme used for signing")
    @click.option("--remove-targets/--no-remove-targets", default=True, help="Should targets delegated to this role also be removed. If not removed, they are signed by the parent role")
    @pin_managed
    def remove(role, path, keystore, scheme, remove_targets, no_commit, prompt_for_keys, pin_manager, keys_description, no_remote_check):
        remove_role(
            path=path,
            pin_manager=pin_manager,
            role=role,
            keystore=keystore,
            scheme=scheme,
            remove_targets=remove_targets,
            commit=not no_commit,
            prompt_for_keys=prompt_for_keys,
            keys_description=keys_description,
            skip_remote_check=no_remote_check,
        )
    return remove


def remove_paths_command():
    @click.command(help="""Remove paths from delegated role""")
    @find_repository
    @catch_cli_exception(handle=TAFError)
    @common_repo_edit_options
    @click.option("--path", default=".", help="Authentication repository's location. If not specified, set to the current directory")
    @click.option("--delegated-path", multiple=True, help="A list of paths to be removed")
    @click.option("--scheme", default=DEFAULT_RSA_SIGNATURE_SCHEME, help="A signature scheme used for signing")
    @click.option("--commit-msg", default=None, help="Commit message")
    @pin_managed
    def remove_delegated_paths(path, delegated_path, keystore, scheme, no_commit, prompt_for_keys, pin_manager, keys_description, no_remote_check, commit_msg):
        if not delegated_path:
            print("Specify at least one role")
            return

        remove_paths(
            path=path,
            pin_manager=pin_manager,
            paths=delegated_path,
            keystore=keystore,
            scheme=scheme,
            commit=not no_commit,
            prompt_for_keys=prompt_for_keys,
            keys_description=keys_description,
            skip_remote_check=no_remote_check,
            commit_msg=commit_msg,
        )
    return remove_delegated_paths


def add_signing_key_command():
    @click.command(help="""
        Add a new signing key. This will make it possible to a sign metadata files
        corresponding to the specified roles with another key. Although private keys are
        used for signing, key identifiers are calculated based on the public keys. This
        means that it's necessary to enter the public key in order to register
        a new signing key. Public key can be loaded from a file, in which case it is
        necessary to specify its path as the pub_key parameter's value. If this option
        is not used when calling this command, the key can be directly entered later.
        """)
    @find_repository
    @catch_cli_exception(handle=TAFError)
    @common_repo_edit_options
    @click.option("--path", default=".", help="Authentication repository's location. If not specified, set to the current directory")
    @click.option("--role", multiple=True, help="A list of roles to whose list of signing keys the new key should be added")
    @click.option("--pub-key-path", default=None, help="Path to the public key corresponding to the private key which should be registered as the role's signing key")
    @click.option("--scheme", default=DEFAULT_RSA_SIGNATURE_SCHEME, help="A signature scheme used for signing")
    @click.option("--commit-msg", default=None, help="Commit message")
    @pin_managed
    def adding_signing_key(path, role, pub_key_path, keystore, scheme, no_commit, prompt_for_keys, pin_manager, keys_description, no_remote_check, commit_msg):
        if not role:
            print("Specify at least one role")
            return

        add_signing_key(
            path=path,
            pin_manager=pin_manager,
            roles=role,
            pub_key_path=pub_key_path,
            keystore=keystore,
            scheme=scheme,
            commit=not no_commit,
            prompt_for_keys=prompt_for_keys,
            keys_description=keys_description,
            skip_remote_check=no_remote_check,
            commit_msg=commit_msg,
        )
    return adding_signing_key


def revoke_signing_key_command():
    @click.command(help="""
        Revoke a signing key.
        """)
    @find_repository
    @catch_cli_exception(handle=TAFError)
    @common_repo_edit_options
    @click.argument("keyid")
    @click.option("--path", default=".", help="Authentication repository's location. If not specified, set to the current directory")
    @click.option("--role", multiple=True, help="A list of roles from which to remove the key. If unspecified, the key is removed from all roles by default.")
    @click.option("--scheme", default=DEFAULT_RSA_SIGNATURE_SCHEME, help="A signature scheme used for signing")
    @click.option("--commit-msg", default=None, help="Commit message")
    @pin_managed
    def revoke_key(path, role, keyid, keystore, scheme, no_commit, prompt_for_keys, pin_manager, keys_description, no_remote_check, commit_msg):

        revoke_signing_key(
            path=path,
            pin_manager=pin_manager,
            roles=role,
            key_id=keyid,
            keystore=keystore,
            scheme=scheme,
            commit=not no_commit,
            prompt_for_keys=prompt_for_keys,
            keys_description=keys_description,
            skip_remote_check=no_remote_check,
            commit_msg=commit_msg,
        )
    return revoke_key


def rotate_signing_key_command():
    @click.command(help="""
        Rotate a signing key.
        """)
    @find_repository
    @catch_cli_exception(handle=TAFError)
    @click.argument("keyid")
    @click.option("--path", default=".", help="Authentication repository's location. If not specified, set to the current directory")
    @click.option("--role", multiple=True, help="A list of roles from which to rotate the key. Rotate from all by default")
    @click.option("--pub-key-path", default=None, help="Path to the public key corresponding to the private key which should be registered as the role's signing key")
    @click.option("--scheme", default=DEFAULT_RSA_SIGNATURE_SCHEME, help="A signature scheme used for signing")
    @click.option("--revoke-commit-msg", default=None, help="Revoke key commit message")
    @click.option("--add-commit-msg", default=None, help="Add new signing key commit message")
    @common_repo_edit_options
    @pin_managed
    def rotate_key(path, role, keyid, pub_key_path, keystore, scheme, prompt_for_keys, revoke_commit_msg, add_commit_msg, pin_manager, keys_description, no_remote_check, no_commit):
        rotate_signing_key(
            path=path,
            pin_manager=pin_manager,
            roles=role,
            key_id=keyid,
            keystore=keystore,
            scheme=scheme,
            prompt_for_keys=prompt_for_keys,
            pub_key_path=pub_key_path,
            revoke_commit_msg=revoke_commit_msg,
            add_commit_msg=add_commit_msg,
            keys_description=keys_description,
            skip_remote_check=no_remote_check,
            commit=not no_commit
        )
    return rotate_key


def list_keys_command():
    @click.command(help="""
        List all keys of the specified role. If certs directory exists and contains certificates exported from YubiKeys,
        include additional information read from these certificates, like name or organization.
        """)
    @find_repository
    @catch_cli_exception(handle=TAFError)
    @click.argument("role")
    @click.option("--path", default=".", help="Authentication repository's location. If not specified, set to the current directory")
    def list_keys(role, path):
        key_infos = list_keys_of_role(
            path=path,
            role=role,
        )
        print("\n".join(key_infos))
    return list_keys


def attach_to_group(group):

    group.add_command(add_roles_command(), name='add')
    group.add_command(add_role_paths_command(), name='add-role-paths')
    group.add_command(remove_paths_command(), name='remove-paths')
    # group.add_command(remove_role_command(), name='remove')
    group.add_command(add_signing_key_command(), name='add-signing-key')
    group.add_command(revoke_signing_key_command(), name='revoke-key')
    group.add_command(rotate_signing_key_command(), name='rotate-key')
    group.add_command(list_keys_command(), name='list-keys')
    group.add_command(export_roles_description_command(), name="export-description")
